import matplotlib

matplotlib.use("Agg")
import numpy as np
import torch, os, psutil, pathlib, sys
import torch.nn as nn
from lightningcast import torch_models
import pickle
import matplotlib.pyplot as plt
from lightningcast import performance_diagrams
from lightningcast import roc_curves
from lightningcast import attributes_diagrams
from sklearn.metrics import mean_squared_error, roc_auc_score
import sklearn.metrics
from lightningcast.sat_zen_angle import sat_zen_angle
from lightningcast import utils
import glob
from lightningcast import torch_generator
import argparse
import time
import zarr

FONT_SIZE = 8
plt.rc("font", size=FONT_SIZE)
plt.rc("axes", titlesize=FONT_SIZE)
plt.rc("axes", labelsize=FONT_SIZE)
plt.rc("xtick", labelsize=FONT_SIZE)
plt.rc("ytick", labelsize=FONT_SIZE)
plt.rc("legend", fontsize=FONT_SIZE)
plt.rc("figure", titlesize=FONT_SIZE)
# --------------------------------------------------------------------------------------------------------------------------------------------------#
def calc_verification_scores(preds, test_labels):
    model_auc = roc_auc_score(test_labels, preds)
    model_brier_score = mean_squared_error(test_labels, preds)
    climo_brier_score = mean_squared_error(
        test_labels,
        np.ones(test_labels.size, dtype=np.float32)
        * test_labels.sum()
        / test_labels.size,
    )
    model_brier_skill_score = 1 - model_brier_score / climo_brier_score
    print(f"AUC: {model_auc:0.5f}")
    print(f"Brier Score: {model_brier_score:0.3f}")
    print(f"Brier Score (Climatology): {climo_brier_score:0.3f}")
    print(f"Brier Skill Score: {model_brier_skill_score:0.3f}")

    return {
        "model_auc": model_auc,
        "model_brier_score": model_brier_score,
        "climo_brier_score": climo_brier_score,
        "model_brier_skill_score": model_brier_skill_score,
    }


# --------------------------------------------------------------------------------------------------------------------------------------------------#
def get_area_under_perf_diagram(success_ratio_by_threshold, pod_by_threshold):
    """Computes area under performance diagram.
    T = number of binarization thresholds
    :param success_ratio_by_threshold: length-T numpy array of success ratios.
    :param pod_by_threshold: length-T numpy array of corresponding POD values.
    :return: area_under_curve: Area under performance diagram.
    Credit: https://github.com/thunderhoser/GewitterGefahr/blob/master/gewittergefahr/gg_utils/model_evaluation.py
    """

    num_thresholds = len(success_ratio_by_threshold)
    expected_dim = np.array([num_thresholds], dtype=int)

    sort_indices = np.argsort(success_ratio_by_threshold)
    success_ratio_by_threshold = success_ratio_by_threshold[sort_indices]
    pod_by_threshold = pod_by_threshold[sort_indices]

    nan_flags = np.logical_or(
        np.isnan(success_ratio_by_threshold), np.isnan(pod_by_threshold)
    )
    if np.all(nan_flags):
        return np.nan

    real_indices = np.where(np.invert(nan_flags))[0]

    return sklearn.metrics.auc(
        success_ratio_by_threshold[real_indices], pod_by_threshold[real_indices]
    )


# --------------------------------------------------------------------------------------------------------------------------------------------------#
def verification(conv_preds, test_out, outdir, second_preds_dict=None, second_sd=None):

    print("calculating scores...")
    t0 = time.time()
    scores = calc_verification_scores(conv_preds, test_out)
    of = open(os.path.join(outdir, "verification_scores.txt"), "w")
    for k, v in scores.items():
        of.write(k + ": " + "{:.5f}".format(v) + "\n")
    # of.write('Binary accuracy: ' + "{:.3f}".format(keras_metrics.accuracy(test_out, conv_preds)) + '\n')
    of.write("Test sample size: " + str(len(test_out)) + "\n")
    print("done with scores", int(time.time() - t0))

    # performance diagram
    t1 = time.time()
    scores_dict = performance_diagrams.plot_performance_diagram(
        observed_labels=test_out, forecast_probabilities=conv_preds
    )
    csi = scores_dict["csi"]
    pss = scores_dict["pss"]
    bin_acc = scores_dict["bin_acc"]
    bin_bias = scores_dict["bin_bias"]
    sr = scores_dict["sr"]
    pod = scores_dict["pod"]
    aupd = get_area_under_perf_diagram(sr, pod)
    print("perf diagram", int(time.time() - t1))

    if second_sd:
        new_sd = pickle.load(open(second_sd, "rb"))
    if second_preds_dict:
        second_preds = pickle.load(open(second_preds_dict, "rb"))

    perf_diagram_file_name = os.path.join(outdir, "performance_diagram.png")

    #  if(second_sd): plot_line(plt.gca(), new_sd, perf=True)

    plt.savefig(perf_diagram_file_name, dpi=300, bbox_inches="tight")
    plt.close()

    # save CSI
    best_ind = np.argmax(csi)
    prob_of_max_csi = (
        best_ind / 100.0
    )  # this only works in binarization_thresholds is linspace(0,1.01,0.001)
    of.write("AUPD: " + "{:.4f}".format(aupd) + "\n")
    of.write(
        "Max CSI: "
        + "{:.4f}".format(np.max(csi))
        + "; at prob: "
        + "{:.3f}".format(prob_of_max_csi)
        + "\n"
    )
    of.write(
        "bin. freq. bias (at max CSI): " + "{:.3f}".format(bin_bias[best_ind]) + "\n"
    )
    of.write("bin. acc. (at max CSI): " + "{:.4f}".format(bin_acc[best_ind]) + "\n")
    of.write(
        "Max PSS: "
        + "{:.4f}".format(np.max(pss))
        + "; at prob: "
        + "{:.3f}".format(np.argmax(pss) / 100.0)
        + "\n"
    )
    of.close()

    # ROC curve
    t2 = time.time()
    pofd_by_threshold, pod_by_threshold = roc_curves.plot_roc_curve(
        observed_labels=test_out, forecast_probabilities=conv_preds
    )
    scores_dict["pofd"] = pofd_by_threshold
    print("roc_curve", int(time.time() - t2))

    #  if(second_sd): plot_line(plt.gca(), new_sd, roc=True)

    area_under_roc_curve = scores["model_auc"]
    title_string = "Area under ROC curve: {0:.4f}".format(area_under_roc_curve)
    plt.title(title_string)
    roc_curve_file_name = os.path.join(outdir, "roc_curve.png")
    plt.savefig(roc_curve_file_name, dpi=300, bbox_inches="tight")
    plt.close()

    plot_hist = False if (second_preds_dict) else True

    # attributes diagram
    t3 = time.time()
    (
        mean_forecast_probs,
        mean_event_frequencies,
        num_examples_by_bin,
        ax_ob,
        fig_ob,
    ) = attributes_diagrams.plot_attributes_diagram(
        observed_labels=test_out,
        forecast_probabilities=conv_preds,
        num_bins=20,
        return_main_axes=True,
        nboots=0,
        plot_hist=plot_hist,
    )
    print("att diagram", time.time() - t3)

    #  if(second_preds_dict): plot_line(ax_ob, fig_ob, second_preds, atts=True)
    #  if(second_sd): plot_line(ax_ob, new_sd, atts=True)

    attr_diagram_file_name = os.path.join(outdir, "attributes_diagram.png")
    plt.savefig(attr_diagram_file_name, dpi=300, bbox_inches="tight")
    plt.close()

    scores_dict["mean_forecast_probs"] = mean_forecast_probs
    scores_dict["mean_event_frequencies"] = mean_event_frequencies
    scores_dict["num_examples_by_bin"] = num_examples_by_bin

    return scores_dict


# --------------------------------------------------------------------------------------------------------------------------------------------------#
if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "indir",
        help="Directory containing trained CNN model and config.",
        type=str,
        nargs=1,
    )
    parser.add_argument(
        "-o",
        "--outdir",
        help="Output directory for model and figures. Default=args.indir",
        nargs=1,
        type=str,
    )
    parser.add_argument(
        "-s",
        "--stride",
        help="Number of pixels to sride in x and y directions, to save time. Default=4",
        nargs=1,
        type=int,
        default=[4],
    )
    parser.add_argument(
        "-n",
        "--ngpus",
        help="Number of GPUs to use. Default = 1. Batch size will be ngpus. Set to 0 for CPU-only.",
        nargs=1,
        type=int,
        default=[1],
    )

    args = parser.parse_args()

    stride = args.stride[0]
    indir = args.indir[0]
    ngpus = args.ngpus[0]
    config = pickle.load(open(os.path.join(indir, "model_config.pkl"), "rb"))
    outdir = args.outdir[0] if (args.outdir) else args.indir[0]
    utils.mkdir_p(outdir)

    # os.environ["CUDA_DEVICE_ORDER"]="PCI_BUS_ID";
    # os.environ["CUDA_VISIBLE_DEVICES"]="2"

    # static reduced-res (RR) lats, lons, satzen
    scalar_vars = config["scalar_vars"]
    if (
        "lat" in scalar_vars
        or "lon" in scalar_vars
        or "satzen" in scalar_vars
        or "solzen" in scalar_vars
    ):
        _, _, _, lon, lat = utils.get_area_definition(
            "static/GOES_East.nc", return_latlons=True
        )
        srrf = config["scalar_reduce_res_fac"]
        lonRR = lon[srrf // 2 :: srrf, srrf // 2 :: srrf]
        latRR = lat[srrf // 2 :: srrf, srrf // 2 :: srrf]
        satzenRR = (
            sat_zen_angle(xlat=latRR, xlon=lonRR, satlat=0, satlon=-75.0)
            if ("satzen" in scalar_vars)
            else None
        )
    else:
        latRR = lonRR = satzenRR = None

    torch_model = getattr(torch_models, config["cnn"])
    model = torch_model(config)
    if ngpus > 0 and torch.cuda.is_available():
        device = torch.device("cuda")
        batch_size = ngpus  # * 2
    else:
        ngpus = 0
        device = torch.device("cpu")
        batch_size = 1

    best_model = np.sort(glob.glob(os.path.join(indir, "model*.pt")))[-1]
    print("opening " + best_model)
    mdict = torch.load(best_model, map_location=device)
    model.load_state_dict(mdict["model_state_dict"])
    ngpu = torch.cuda.device_count()

    if ngpus > 1:
        print("Let's use", ngpus, "GPUs!")
        model = nn.DataParallel(model)
    model.to(device)
    model.eval()  # Needed to set dropout and batch normalization layers to evaluation mode before running inference.
    # Failing to do this will yield inconsistent inference results.
    print(f"Using device = {device}")

    DG = getattr(torch_generator, config["datagen"])

    if config["datagen"] == "DatasetZARR_UNET":
        zf = zarr.open(config["data"])
        # get the inds
        part_file = pickle.load(open(os.path.join(indir, "part_labs.pkl"), "rb"))
        partition = part_file["partition"]
        labels = part_file["labels"]
        val_ind = partition["validation"]
        val_mask = part_file["val_mask"]
        n_vsamples = len(val_ind)
        validation_set = DG(val_ind, config, zarr=zf)
    # if we're using a list of dates
    elif len(config["val_dates"]):
        val_files = []
        for vd in config["val_dates"]:
            val_files += glob.glob(os.path.join(config["data"], str(vd), "*.netcdf"))
        val_files = np.sort(val_files)
        n_vsamples = len(val_files)
        validation_set = DG(val_files, config, lats=latRR, lons=lonRR, satzen=satzenRR)
    else:
        # We're using samplelists (pickles containing lists of dicts)
        val_list = pickle.load(open(config["val_list"], "rb"))
        n_vsamples = len(val_list)
        validation_set = DG(val_list, config, lats=latRR, lons=lonRR, satzen=satzenRR)

    ny = config["input_tuples"][-1][-2]
    nx = config["input_tuples"][-1][-1]

    nclass = config["nclass"]
    if config["nclass"] > 1:  # multi-label classification
        preds = np.empty((n_vsamples, nclass, ny, nx), dtype=np.float16)
        labels = np.empty((n_vsamples, ny, nx), dtype=np.float16)
    else:  # binary classification
        preds = np.empty((n_vsamples, 1, ny, nx), dtype=np.float16)
        labels = np.empty((n_vsamples, 1, ny, nx), dtype=np.float16)

    assert batch_size >= ngpus
    remainder = n_vsamples % batch_size

    params = {
        "batch_size": batch_size,  # best to use a batch_size that is a multiple of your ngpus
        "shuffle": False,
        "num_workers": 8,
    }

    validation_generator = torch.utils.data.DataLoader(validation_set, **params)

    # Evaluation
    starttime = time.time()
    with torch.set_grad_enabled(False):
        cter = 0
        #  print('cter',cter)
        for local_batch, local_labels in validation_generator:  # for each sample
            print(cter)
            # Transfer to device (GPU or CPU)
            local_batch_list = []
            for lb in local_batch:  # for each input of each sample
                local_batch_list.append(lb.to(device, dtype=torch.float))

            # make local_labels numpy and add to container array

            # print(labels.shape)
            # print(local_labels.cpu().numpy().shape)
            # print('')
            # print(model(local_batch_list).cpu().numpy().shape)
            # print(preds.shape)

            try:
                labels[cter : cter + batch_size] = local_labels.cpu().numpy()
            except ValueError:
                labels[cter : cter + remainder] = local_labels.cpu().numpy()

            # Model computations on input data and add to container array
            outputs = model(local_batch_list).cpu().numpy()
            outputs[outputs > 1] = 1

            try:
                preds[cter : cter + batch_size] = outputs
            except ValueError:
                preds[cter : cter + remainder] = outputs
            cter += 1

        del model, outputs, local_batch_list, local_batch, local_labels
    # there seems to be some memory leak making predictions in this way, but I'm not sure why.

    print(
        "Made predictions in " + "{:.0f}".format(time.time() - starttime) + " seconds"
    )
    torch.cuda.empty_cache()  # doesn't actually release GPU RAM. Some variable still referencing it?

    if nclass > 1:
        print(
            f"Assuming {nclass} classes, with positive class as the last index of dim=1..."
        )
        preds = preds[:, -1, ...]

    try:
        preds = np.squeeze(preds, axis=1)
    except ValueError:
        pass  # assumes we're doing multi-classification
    preds = preds[:, ::stride, ::stride]  # stride in both x and y
    preds = preds.flatten().astype(
        np.float32
    )  # float32 precision needed for validation routines (e.g., MSE)

    try:
        labels = np.squeeze(labels, axis=1)
    except ValueError:
        pass  # assumes we're doing multi-classification
    labels = labels[:, ::stride, ::stride]  # stride in both x and y
    labels = labels.flatten().astype(
        np.float32
    )  # float32 precision needed for validation routines (e.g., MSE)

    if (
        preds.nbytes / (1024**3) > 15
    ):  # If data > 10 GB, just score probs that are > than some threshold.
        # We do this because in the verification function, we seem to run out
        # of RAM, even with 180GB, if the data are 15GB or more.
        good_ind = np.where(preds > 0.001)
        print(len(preds))
        preds = preds[good_ind]
        print(len(preds))
        labels = labels[good_ind]

    # binarize
    print("binarizing...")
    labels[labels >= config["binarize"]] = 1
    labels[labels < config["binarize"]] = 0

    scores_dict = verification(preds, labels, outdir)

    pickle.dump(scores_dict, open(os.path.join(outdir, "scores_dict.pkl"), "wb"))
    of = open(os.path.join(outdir, "verification_scores.txt"), "a")
    try:
        of.write("data source: " + config["data"] + "\n")
    except TypeError:
        pass
    of.close()

# using numpy arrays...all in memory

# with torch.set_grad_enabled(False):
#  for i in range(len(inputs)):
#    inputs[i] = torch.from_numpy(inputs[i].astype(np.float32)).to(device,dtype=torch.float)

# for i in range(len(inputs)):
#  inputs[i] = torch.from_numpy(inputs[i].astype(np.float32))

# output = model(inputs).data.cpu().numpy()


# output = np.squeeze(output,axis=(0,1))
