import json
import os, sys
from datetime import datetime
import argparse
import collections

import inspect

current_dir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
sys.path.insert(0, os.path.dirname(current_dir))

from lightningcast.fire_event_timeseries import make_fire_points
from subprocess import call


parser = argparse.ArgumentParser(
    description="Takes the NIFC centroid geojson and creates a RealEarth compatible geojson of NIFC events."
)
parser.add_argument(
    "nifc_geojson", help="Full path and filename for NIFC geojson.", type=str, nargs=1
)
parser.add_argument(
    "-ru",
    "--re_upload",
    help="Full path to re_upload executable. No upload will occur if not provided.",
    default=[None],
    type=str,
    nargs=1,
)


args = parser.parse_args()
re_upload = args.re_upload[0]

# Open NIFC geojson
f = open(args.nifc_geojson[0])
point_data = json.load(f)
f.close()


# Collect fire event data
# Assumes you want every event
epoch_target = epoch_target1 = int(
    (datetime.utcnow() - datetime(1970, 1, 1, 0, 0)).total_seconds()
)
epoch_target0 = 0
update_threshold = 9999  # days
max_incident_lon = 179.99
min_incident_lon = -179.99
max_incident_lat = 90
min_incident_lat = -90
ipoints, _ = make_fire_points(
    point_data,
    epoch_target,
    epoch_target0,
    epoch_target1,
    update_threshold,
    min_incident_lon,
    max_incident_lon,
    min_incident_lat,
    max_incident_lat,
)

# Create output geojson

outdict = collections.OrderedDict()
outdict["type"] = "FeatureCollection"
flist = []
dashboard = "https://grafana-public.ssec.wisc.edu/grafana/d/mr66nbemzLIGHTNING/nifc-wildfire-lightning-dashboard?orgId=1&refresh=5m&var-wildfires"

for i in ipoints:
    last_update_dt = datetime.utcfromtimestamp(int(i["epoch_last_update"]))
    info = f'<a target="_blank" href="{dashboard}={i["incident_name"]}_{i["incident_county"].upper()}_{i["incident_state"].upper()}">{i["incident_name"]}</a>'
    if "AK" in i["incident_state"].upper():
        county_type = "Borough"
    elif "LA" in i["incident_state"].upper():
        county_type = "Parish"
    else:
        county_type = "County"
    info += f"<br>{i['incident_county']} {county_type}, {i['incident_state'][-2:]}"
    info += f"<br>Last NIFC update: {last_update_dt.strftime('%Y-%m-%d %H:%M UTC')}"

    fdict = collections.OrderedDict({"type": "Feature"})
    fdict["geometry"] = collections.OrderedDict(
        {"type": "Point", "coordinates": [i["incident_lon"], i["incident_lat"]]}
    )
    fdict["properties"] = collections.OrderedDict(
        {"COLOR": "252 114 0", "OPACITY": "100"}
    )
    fdict["properties"]["INFO"] = info
    fdict["properties"]["WIDTH"] = "8"
    flist.append(fdict)


# Write out the RealEarth geojson

outdict["features"] = flist
outjson_obj = json.dumps(outdict, indent=2)

ts = datetime.utcnow().strftime("%Y%m%d_%H%M%S")
outfile = f"CURRENTNIFC_{ts}.force.json"
g = open(outfile, "w")
g.write(outjson_obj)
g.close()


# Send to RealEarth VM

if (
    re_upload
):  # re_upload is either None or a string to full path of the re_upload script
    s = call([re_upload, outfile])
    # Remove file
    os.remove(outfile)
