import matplotlib.pyplot as plt
import numpy as np
import glob
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from netCDF4 import Dataset
import argparse
from datetime import datetime, timedelta
import sys, os
from lightningcast import utils
import pyresample as pr
from lightningcast.predict_ltg import get_rgb_colors, read_netcdf
from pyproj import Proj

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"


def rand_cmap(
    nlabels, type="bright", first_color_black=True, last_color_black=False, verbose=True
):
    """
    Creates a random colormap to be used together with matplotlib. Useful for segmentation tasks
    :param nlabels: Number of labels (size of colormap)
    :param type: 'bright' for strong colors, 'soft' for pastel colors
    :param first_color_black: Option to use first color as black, True or False
    :param last_color_black: Option to use last color as black, True or False
    :param verbose: Prints the number of labels and shows the colormap. True or False
    :return: colormap for matplotlib
    """
    from matplotlib.colors import LinearSegmentedColormap
    import colorsys
    import numpy as np

    if type not in ("bright", "soft"):
        print('Please choose "bright" or "soft" for type')
        return

    if verbose:
        print("Number of labels: " + str(nlabels))

    # Generate color map for bright colors, based on hsv
    if type == "bright":
        randHSVcolors = [
            (
                np.random.uniform(low=0.0, high=1),
                np.random.uniform(low=0.2, high=1),
                np.random.uniform(low=0.9, high=1),
            )
            for i in range(nlabels)
        ]

        # Convert HSV list to RGB
        randRGBcolors = []
        for HSVcolor in randHSVcolors:
            randRGBcolors.append(
                colorsys.hsv_to_rgb(HSVcolor[0], HSVcolor[1], HSVcolor[2])
            )

        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]

        random_colormap = LinearSegmentedColormap.from_list(
            "new_map", randRGBcolors, N=nlabels
        )

    # Generate soft pastel colors, by limiting the RGB spectrum
    if type == "soft":
        low = 0.6
        high = 0.95
        randRGBcolors = [
            (
                np.random.uniform(low=low, high=high),
                np.random.uniform(low=low, high=high),
                np.random.uniform(low=low, high=high),
            )
            for i in range(nlabels)
        ]

        if first_color_black:
            randRGBcolors[0] = [0, 0, 0]

        if last_color_black:
            randRGBcolors[-1] = [0, 0, 0]
        random_colormap = LinearSegmentedColormap.from_list(
            "new_map", randRGBcolors, N=nlabels
        )

    # Display colorbar
    if verbose:
        from matplotlib import colors, colorbar
        from matplotlib import pyplot as plt

        fig, ax = plt.subplots(1, 1, figsize=(15, 0.5))

        bounds = np.linspace(0, nlabels, nlabels + 1)
        norm = colors.BoundaryNorm(bounds, nlabels)

        cb = colorbar.ColorbarBase(
            ax,
            cmap=random_colormap,
            norm=norm,
            spacing="proportional",
            ticks=None,
            boundaries=bounds,
            format="%1i",
            orientation="horizontal",
        )

    return random_colormap


#####################################################################################################################
if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Makes images of lightning probability objects."
    )
    parser.add_argument(
        "ltg_obj_dir", help="Directory containing merged objs.", type=str, nargs=1
    )
    parser.add_argument(
        "ltg_prob_dir", help="Directory containing P(LTG) netcdfs.", type=str, nargs=1
    )
    parser.add_argument(
        "outdir", help="output directory for images.", type=str, nargs=1
    )
    parser.add_argument(
        "-g",
        "--glmdir",
        help="root glm dir. Default: /apollo/grain/saved_probsevere_data/lightning/ \
                      ...will append yyyymmdd/GLM/goes_east/agg/",
        type=str,
        nargs=1,
        default=["/apollo/grain/saved_probsevere_data/lightning/"],
    )

    args = parser.parse_args()

    outdir = args.outdir[0]
    utils.mkdir_p(outdir)
    ltg_obj_dir = args.ltg_obj_dir[0]
    ltg_prob_dir = args.ltg_prob_dir[0]
    glmdir = args.glmdir[0]

    ltg_prob_files = np.sort(glob.glob(ltg_prob_dir + "/*.netcdf"))

    state_borders = cfeature.NaturalEarthFeature(
        category="cultural",
        name="admin_1_states_provinces_lakes",
        scale="50m",
        facecolor="none",
    )

    # hardcoding ABI dir
    abidir = "/arcdata/goes/grb/goes16/"

    # get irlats and irlons
    # static 2km lats, lons
    conus2km = Dataset("static/GOES_East.nc", "r")
    satheight = conus2km.variables["goes_imager_projection"].getncattr(
        "perspective_point_height"
    )
    lon0 = conus2km.variables["goes_imager_projection"].getncattr(
        "longitude_of_projection_origin"
    )
    sweep = conus2km.variables["goes_imager_projection"].getncattr("sweep_angle_axis")
    # projections
    p = Proj(proj="geos", h=satheight, lon_0=lon0, sweep=sweep)
    geoproj = ccrs.Geostationary(
        central_longitude=lon0, sweep_axis=sweep, satellite_height=satheight
    )

    # The projection x and y coords equals the scanning angle (in radians) multiplied by the satellite height (http://proj4.org/projections/geos.html)
    irx = conus2km.variables["x"][:] * satheight
    iry = conus2km.variables["y"][:] * satheight
    irxx, iryy = np.meshgrid(irx, iry)
    irlonsfull, irlatsfull = p(irxx, iryy, inverse=True)
    conus2km.close()

    # contour info
    colors = ["#0000FF", "#00FFFF", "#00FF00", "#CC00FF"]
    rgb_colors = get_rgb_colors(colors)
    values = [10, 25, 50, 75]
    linewidths = [0.75, 1, 1, 1]

    ct = 0
    for lpf in ltg_prob_files:
        ncprob = Dataset(lpf, "r")

        # only one var in ncprob
        for key in ncprob.variables.keys():
            varname = key

        probdata = ncprob.variables[varname][:]
        nlat, nlon = probdata.shape

        NWlat = ncprob.Latitude
        NWlon = ncprob.Longitude
        dx = ncprob.LonGridSpacing
        dy = ncprob.LatGridSpacing
        timestamp = ncprob.datetime

        lats = np.zeros((nlat, nlon))
        for ii in range(nlat):
            lats[ii, :] = NWlat - ii * dy
        lons = np.zeros((nlat, nlon))
        for ii in range(nlon):
            lons[:, ii] = NWlon + ii * dx

        ncprob.close()

        # get corresponding object file
        basefile = os.path.basename(lpf)
        objfile = os.path.join(ltg_obj_dir, basefile)

        try:
            ncobj = Dataset(objfile, "r")
        except FileNotFoundError:
            continue
        objs = ncobj.variables["merged_prob_objs"][:]
        ncobj.close()
        print(lpf)

        lonmin = lons[0, 0]
        lonmax = lons[0, -1]
        latmin = lats[-1, 0]
        latmax = lats[0, 0]

        # get ABI crop info
        xmin, ymin = p(lonmin, latmin)
        xmax, ymax = p(lonmax, latmax)
        xmin_ind, xmin = utils.find_nearest(irx, xmin)
        xmax_ind, xmax = utils.find_nearest(irx, xmax)
        ymin_ind, ymin = utils.find_nearest(iry, ymin)
        ymax_ind, ymax = utils.find_nearest(iry, ymax)
        startX = xmin_ind
        endX = xmax_ind
        startY = ymax_ind
        endY = ymin_ind
        dt = datetime.strptime(timestamp, "%Y%m%d-%H%M%S UTC")

        # get ABI data
        CH13_file_patt = dt.strftime(
            os.path.join(
                abidir,
                "%Y",
                "%Y_%m_%d_%j",
                "abi",
                "L1b",
                "RadC",
                "OR_ABI-*C13_G16_s%Y%j%H%M*.nc",
            )
        )
        CH13_file = glob.glob(CH13_file_patt)[0]
        nc = Dataset(CH13_file, "r")
        rad = nc.variables["Rad"][:]
        BTtmp = (
            nc.variables["planck_fk2"][:]
            / (np.log((nc.variables["planck_fk1"][:] / rad) + 1))
            - nc.variables["planck_bc1"][:]
        ) / nc.variables["planck_bc2"][:]
        nc.close()

        # remap the ABI data
        BT = BTtmp[startY:endY, startX:endX]
        irlats = irlatsfull[startY:endY, startX:endX]
        irlons = irlonsfull[startY:endY, startX:endX]
        radar_grid = pr.geometry.GridDefinition(lons=lons, lats=lats)
        sat_grid = pr.geometry.GridDefinition(lons=irlons, lats=irlats)
        remapped_BT = pr.kd_tree.resample_nearest(
            sat_grid, BT, radar_grid, radius_of_influence=4000, fill_value=-999
        )

        # get GLM data
        glmfile_patt = dt.strftime(
            os.path.join(
                glmdir, "%Y%m%d", "GLM", "goes_east", "agg", "%Y%m%d-%H%M*.netcdf"
            )
        )
        try:
            glmfile = glob.glob(glmfile_patt)[0]
        except IndexError:
            for mm in [1, 2, 3, 4, 5]:
                newdt = dt - timedelta(minutes=mm)
                glmfile = glob.glob(
                    newdt.strftime(
                        os.path.join(
                            glmdir,
                            "%Y%m%d",
                            "GLM",
                            "goes_east",
                            "agg",
                            "%Y%m%d-%H%M*.netcdf*",
                        )
                    )
                )
                if glmfile:
                    break
            if glmfile:
                glmdata = {"flash_extent_density": -1}
                ss = read_netcdf(glmfile, data=glmdata)
                fed = glmdata["flash_extent_density"][startY:endY, startX:endX]
            else:
                print("Can't find a good GLM file.")
                sys.exit(1)
        else:
            glmdata = {"flash_extent_density": -1}
            ss = read_netcdf(glmfile, data=glmdata)
            fed = glmdata["flash_extent_density"][startY:endY, startX:endX]

        # remap the GLM data
        remapped_FED = pr.kd_tree.resample_nearest(
            sat_grid, fed, radar_grid, radius_of_influence=4000, fill_value=-999
        )

        # make image
        fig = plt.figure(figsize=(14, 7))

        ax1 = plt.subplot(1, 2, 1, projection=ccrs.PlateCarree())
        ax1.set_extent([lonmin, lonmax, latmin, latmax], ccrs.PlateCarree())
        ax1.add_feature(cfeature.BORDERS)
        ax1.add_feature(state_borders, edgecolor="yellow")
        fs = 14
        # img = ax1.pcolormesh(lons, lats, probdata, vmin=0, vmax=75, cmap=plt.get_cmap('inferno'))
        # cbar = plt.colorbar(img, orientation='horizontal',pad=0.01, extendfrac=0, drawedges=0, ax=ax1,ticks=np.arange(0,76,15))
        # cbar.set_label('P(LTG) in next 60 min [%]',fontsize=fs)
        img = ax1.pcolormesh(
            lons, lats, remapped_BT, vmin=180, vmax=310, cmap=plt.get_cmap("Greys")
        )
        # cbar = plt.colorbar(img, orientation='horizontal',pad=0.01, extendfrac=0, drawedges=0, ax=ax1,ticks=np.arange(200,300,40))
        # cbar.set_label('ABI CH13 BT [K]',fontsize=fs)
        img = ax1.pcolormesh(
            lons,
            lats,
            np.ma.masked_less(remapped_FED, 0.1),
            vmin=0,
            vmax=70,
            cmap=plt.get_cmap("autumn_r"),
        )
        cbar = plt.colorbar(
            img,
            orientation="horizontal",
            pad=0.01,
            extendfrac=0,
            drawedges=0,
            ax=ax1,
            ticks=np.arange(0, 71, 10),
        )
        cbar.set_label("GLM FED [fl/(5 min)]", fontsize=fs)
        cbar.ax.tick_params(labelsize=fs)
        plt.title("GLM FED and CH13 BT " + timestamp, fontsize=fs)

        # contours
        contour_img = ax1.contour(
            lons, lats, probdata, values, colors=colors, linewidths=linewidths
        )

        ax2 = plt.subplot(1, 2, 2, projection=ccrs.PlateCarree())
        ax2.set_extent([lonmin, lonmax, latmin, latmax], ccrs.PlateCarree())
        ax2.add_feature(cfeature.BORDERS)
        ax2.add_feature(state_borders, edgecolor="yellow")
        if ct == 0:
            uniq_objs = np.unique(objs)
            uniq_objs = uniq_objs[uniq_objs > 0]
            minid = uniq_objs[0]
            maxid = uniq_objs[-1]
            randcolors = rand_cmap(int(maxid + 5000 - minid), verbose=False)

        img = ax2.pcolormesh(
            lons, lats, objs + 0.5, vmin=minid - 1, vmax=maxid + 5000, cmap=randcolors
        )

        cbar = plt.colorbar(
            img, orientation="horizontal", pad=0.01, extendfrac=0, drawedges=0, ax=ax2
        )
        cbar.set_label("object ID []", fontsize=fs)
        cbar.set_ticks([])
        plt.title("Probability objects " + timestamp, fontsize=fs)

        plt.savefig(
            os.path.join(outdir, timestamp.split(" ")[0] + ".png"), bbox_inches="tight"
        )
        plt.close()
        ct += 1
        # sys.exit()
