import zarr
import numpy as np
import pandas as pd
import xarray as xr
import glob
import os, sys
from datetime import datetime, timedelta
import time
from multiprocessing import Pool
from netCDF4 import Dataset
from collections import OrderedDict
import itertools

os.environ["HDF5_USE_FILE_LOCKING"] = "FALSE"
sys.path.insert(0, "..")
from solar_angles import solar_angles

t0 = time.time()


def get_data(ncs, seq_len):

    recs = {"ABI_date": [], "ABI_time": [], "ABI_datetime": [], "FED": []}

    for s in ncs:
        variables = {"flash_extent_density": []}

        print(s)
        # 20180611-001719.netcdf
        basefile = os.path.basename(s)
        #  dirname = os.path.dirname(s)
        #  dirbase = os.path.dirname(dirname) #remove yyyymmdd
        #  parts = basefile.split('_')
        #  dt = datetime.strptime(parts[-1][0:15],'%Y%m%d-%H%M%S')
        #  prev_dts = [dt - timedelta(minutes=sl*5) for sl in range(seq_len)]
        #  prev_dts.reverse()

        try:
            nc = Dataset(s, "r")
        except (IndexError, OSError):
            continue
        else:
            nc.set_auto_mask(False)

            good_ltg = False
            for var in variables:
                # QC if/elif blocks
                data = nc.variables[var][:]
                if (
                    np.max(data) > 0.1
                ):  # make sure we have some ltg in it. Should we use a fraction instead??
                    good_ltg = True
                    data = np.zeros((1500, 2500))
                    dat = nc.variables[var][:]
                    try:
                        data[:] = dat
                    except ValueError:
                        data[0:1499, 0:2499] = dat
                    variables[var].append(data)

            nc.close()
            if good_ltg:
                recs["ABI_date"].append(int(basefile[0:8]))
                recs["ABI_time"].append(int(basefile[9:13]))
                # recs['ABI_datetime'].append(datetime.strptime(basefile,'%Y%m%d-%H%M%S.netcdf'))
                for var in variables:
                    variables[var] = np.stack(variables[var], axis=-1)
                    if var == "flash_extent_density":
                        recs["FED"].append(variables[var])

    return recs


##################################################################################################################

if __name__ == "__main__":
    date = sys.argv[1]
    topdt = datetime.strptime(str(date), "%Y%m%d")

    nthreads = 6
    chunksize = 1  # number of images/samples to chunk or read at once
    seq_len = 1

    ncdir = "/apollo/grain/saved_probsevere_data/lightning/"
    zname = "/apollo/grain/jcintineo/AI/OCONUS/seg/GLM.zarr"
    abidir = "/arcdata/goes/grb/goes16/"

    for hh in [
        "0[0-4]",
        "0[5-9]",
        "1[0-4]",
        "1[5-9]",
        "2",
    ]:  # this is to keep the RAM down
        pool = Pool(nthreads)

        satfiles = glob.glob(
            topdt.strftime(
                abidir
                + "/%Y/%Y_%m_%d_%j/abi/L1b/RadC/OR_ABI-L1b-RadC-M*C13_G16_s%Y%j"
                + hh
                + "*"
            )
        )
        satfiles = np.sort(satfiles)
        satdts = np.array(
            [
                datetime.strptime(
                    os.path.basename(x).split("_s")[1][0:13], "%Y%j%H%M%S"
                )
                for x in satfiles
            ]
        )

        allncs = []
        for sdt in satdts:
            try:
                f = glob.glob(
                    sdt.strftime(
                        os.path.join(
                            ncdir,
                            "%Y%m%d",
                            "GLM",
                            "goes_east",
                            "agg",
                            "%Y%m%d-%H%M*.netcdf",
                        )
                    )
                )[0]
            except IndexError:
                for ii, mm in enumerate([1, 2, 3]):
                    tmpdt = sdt - timedelta(minutes=mm)
                    try:
                        f = glob.glob(
                            tmpdt.strftime(
                                os.path.join(
                                    ncdir,
                                    "%Y%m%d",
                                    "GLM",
                                    "goes_east",
                                    "agg",
                                    "%Y%m%d-%H%M*.netcdf",
                                )
                            )
                        )[0]
                    except IndexError:
                        pass
                    else:
                        allncs.append(f)
            else:
                allncs.append(f)

        if len(allncs) == 0:
            continue

        interval = int(np.ceil(len(allncs) / nthreads))
        list_of_lists = []
        for n in range(nthreads):
            list_of_lists.append(allncs[n * interval : n * interval + interval])

        results = pool.starmap(get_data, zip(list_of_lists, itertools.repeat(seq_len)))
        pool.close()
        pool.join()

        fields = OrderedDict()
        fieldnames = ["ABI_date", "ABI_time", "FED"]
        for f in fieldnames:
            fields[f] = []

        datasets = {
            "FED": {
                "atts": {
                    "desc": "GLM flash extent density, accumulated over 5 min",
                    "units": "fl",
                    "mean": 0.3536,
                    "std": 2.341,
                }
            },
            "ABI_date": {"atts": {}},
            "ABI_time": {"atts": {}},
        }

        # write zarr
        ds_exists = True if (os.path.isdir(zname)) else False
        root = zarr.open(zname, mode="a")

        for f in fields:
            # if(f == 'solzen'):
            #  dts = [datetime.strptime(str(x)+str(y).zfill(4),'%Y%m%d%H%M') for x,y in zip(fields['ABI_date'],fields['ABI_time'])]
            #  solzen,solaz = solar_angles(dt=dts,lon=fields['lon'],lat=fields['lat'])
            #  fields[f] = solzen
            # else:
            for ii in range(
                len(results)
            ):  # concatenate list of results (dicts) and then stck them into an array
                fields[f] += results[ii][f]
            fields[f] = np.stack(fields[f])
            if f == "FED":
                ind = fields[f] > 9999
                fields[f][ind] = 0
            datasets[f]["data"] = fields[f]

            print(fields[f].shape)

            # set chunksizes
            if len(datasets[f]["data"].shape) == 1:
                chunks = (10000,)
            else:
                chunks = (
                    chunksize,
                    datasets[f]["data"].shape[1],
                    datasets[f]["data"].shape[2],
                    1,
                )

            # create data
            if ds_exists:  # just append along time axis
                newarr = zarr.zeros(
                    datasets[f]["data"].shape,
                    chunks=chunks,
                    dtype=str(datasets[f]["data"].dtype),
                )
                newarr[:] = datasets[f]["data"]
                root[f].append(newarr, axis=0)
            else:  # start of new data array in group
                newarr = root.zeros(
                    f,
                    shape=datasets[f]["data"].shape,
                    chunks=chunks,
                    dtype=str(datasets[f]["data"].dtype),
                )
                newarr[:] = datasets[f]["data"]
                # set attributes
                for att in datasets[f]["atts"]:
                    newarr.attrs[att] = datasets[f]["atts"][att]

    print(time.time() - t0)
