from __future__ import absolute_import, division, print_function, unicode_literals
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model, save_model, load_model
from pyproj import Proj
import sys, os

os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
os.environ["CUDA_VISIBLE_DEVICES"] = "-1"

# sys.path.insert(0,'../static/')
import numpy as np
import numpy.ma as ma
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import zarr
import time
import cartopy.crs as ccrs
import cartopy.feature as cfeature
from netCDF4 import Dataset

state_borders = cfeature.NaturalEarthFeature(
    category="cultural",
    name="admin_1_states_provinces_lakes",
    scale="50m",
    facecolor="none",
)

goesnc = Dataset("../static/GOES_East.nc", "r")
gip = goesnc.variables["goes_imager_projection"]
x = goesnc.variables["x"][:]
y = goesnc.variables["y"][:]

geoproj = ccrs.Geostationary(
    central_longitude=gip.longitude_of_projection_origin,
    sweep_axis=gip.sweep_angle_axis,
    satellite_height=gip.perspective_point_height,
)

x *= gip.perspective_point_height
y *= gip.perspective_point_height
x = x[1510 : 2310 + 160]
y = y[0:800]
xx, yy = np.meshgrid(x, y)
goesnc.close()


zf = "/apollo/grain/jcintineo/AI/OCONUS/seg/unet_train_small.zarr"
model = os.path.join(
    "/home",
    "jcintineo",
    "probsevere_oconus",
    "src",
    "unet",
    "c13_small",
    "fit_conv_model.h5",
)

ds = zarr.open(zf, "r")

ind = [100]

ch13 = ds.CH13.get_orthogonal_selection(
    (ind, slice(None), slice(None), slice(None), slice(None))
)
ch13_norm = (ch13 - ds.CH13.attrs["mean"]) / ds.CH13.attrs["std"]
truth = ds.MAXFED_5to60.get_orthogonal_selection((ind, slice(None), slice(None), 0))

abi_date = ds.ABI_date[ind[0]]
abi_time = ds.ABI_time[ind[0]]
timestamp = str(abi_date) + "-" + str(abi_time).zfill(4)

conv_model = load_model(model, compile=False)

t0 = time.time()
new_preds = conv_model.predict((ch13_norm))
new_preds = new_preds[0]
print("max new_preds:", np.max(new_preds))
truth = truth[0]
print(time.time() - t0)

ch13 = ch13[0, -1, ..., 0]  # this is the t0 frame
irny, irnx = ch13.shape

# get aspect ratio for IR data
default_imgsize = 10  # inches
asp = irny / irnx
if asp > 1:
    fig_width_in = default_imgsize * asp ** (-1)
    fig_height_in = default_imgsize
else:
    fig_width_in = default_imgsize
    fig_height_in = default_imgsize * asp
fig = plt.figure(figsize=(fig_width_in, fig_height_in))

ax = fig.add_axes([0, 0, 1, 1], projection=geoproj)
# set extent of image
xmin = np.min(x)
xmax = np.max(x)
ymin = np.min(y)
ymax = np.max(y)
ax.set_extent([xmin, xmax, ymin, ymax], crs=geoproj)
ax.add_feature(cfeature.BORDERS)
ax.add_feature(state_borders, edgecolor="tan")

irimg = ax.pcolormesh(xx, yy, ch13, vmin=180, vmax=310, cmap=plt.get_cmap("Greys"))
cbaxes = fig.add_axes([0.01, -0.05, 0.27, 0.03])  # left bottom width height
cbar = plt.colorbar(
    irimg,
    orientation="horizontal",
    pad=0.01,
    extendfrac=0,
    drawedges=0,
    ax=ax,
    ticks=np.arange(180, 301, 40),
    cax=cbaxes,
)
cbar.set_label("ABI 10.35-$\mu$m brightness temperature [K]", fontsize=10)

glmimg = ax.pcolormesh(
    xx,
    yy,
    ma.masked_less(truth, 0.99),
    vmin=0.99,
    vmax=80,
    cmap=plt.get_cmap("autumn_r"),
    linewidth=0,
    antialiased=True,
    alpha=0.3,
)
cbaxes2 = fig.add_axes([0.31, -0.05, 0.27, 0.03])  # left bottom width height
cbar2 = plt.colorbar(
    glmimg,
    orientation="horizontal",
    pad=0.01,
    extendfrac=0,
    drawedges=0,
    ax=ax,
    ticks=np.arange(10, 71, 20),
    cax=cbaxes2,
)
cbar2.set_label("Max GLM FED $t_{0}$ + 60 min [fl (5 min)$^{-1}$]", fontsize=10)
cbar2.set_alpha(1.0)
cbar2.draw_all()

colors = ["#0F0080", "#1E00FF", "#00ffff", "#CC00FF"]
values = [0.1, 0.25, 0.5, 0.75]
linewidths = [0.5] * len(values)
labels = [str(int(v * 100)) + "%" for v in values]
contour_img = ax.contour(
    xx, yy, new_preds, values, colors=colors, linewidths=linewidths
)  # , cmap=plt.get_cmap('cividis'))
custom_lines = [
    Line2D([0], [0], color=c, lw=lw * 2) for c, lw in zip(colors, linewidths)
]
leg2 = ax.legend(
    custom_lines,
    labels,
    fontsize=11,
    loc="lower right",
    handlelength=1.5,
    ncol=len(labels),
    bbox_to_anchor=(1, -0.06),
)
ctext = plt.text(
    0.65,
    -0.09,
    "Probability of lightning in 60 min",
    fontsize=14,
    transform=ax.transAxes,
)

ax.set_title(
    (" ").join(["GOES-East", "CONUS", "$t_{0}$", "=", timestamp, "UTC"]), fontsize=16
)

plt.savefig("test2.png", bbox_inches="tight")
