import sys, os
import numpy as np
import collections
import pickle
import glob


def tf_config():
    nframes = 1
    dilation = 1
    shuffle = True
    nworkers = 2  # number of workers for fit_generator for train_set

    minmax = []

    # the values that were used for bytescaling
    bsinfo = {}  # bytescaling info
    bsinfo["CH01"] = {"vmin": 0, "vmax": 1}
    bsinfo["CH02"] = {"vmin": 0, "vmax": 1}
    bsinfo["CH03"] = {"vmin": 0, "vmax": 1}
    bsinfo["CH04"] = {"vmin": 0, "vmax": 0.75}
    bsinfo["CH05"] = {"vmin": 0, "vmax": 0.75}
    bsinfo["CH06"] = {"vmin": 0, "vmax": 0.75}
    bsinfo["CH07"] = {"vmin": 220, "vmax": 360}
    bsinfo["CH08"] = {"vmin": 200, "vmax": 270}
    bsinfo["CH09"] = {"vmin": 200, "vmax": 275}
    bsinfo["CH10"] = {"vmin": 200, "vmax": 280}
    bsinfo["CH11"] = {"vmin": 190, "vmax": 320}
    bsinfo["CH12"] = {"vmin": 190, "vmax": 320}
    bsinfo["CH13"] = {"vmin": 190, "vmax": 320}
    bsinfo["CH14"] = {"vmin": 190, "vmax": 320}
    bsinfo["CH15"] = {"vmin": 190, "vmax": 320}
    bsinfo["CH16"] = {"vmin": 190, "vmax": 300}
    bsinfo["REF10"] = {"vmin": 0, "vmax": 75}
    bsinfo["FED"] = {"vmin": 0, "vmax": 255}
    bsinfo["SATZEN"] = {"vmin": 0, "vmax": 90}
    bsinfo["CLIMO"] = {"vmin": 0, "vmax": 0.23}
    bsinfo["SFC_ELEV"] = {"vmin": -999, "vmax": 3800}
    bsinfo["HRAC_COM_FR"] = {"vmin": 0, "vmax": 0.24}
    bsinfo["LANDMASK"] = {"vmin": 0, "vmax": 1}

    # In native units. For normal floats, substract mean and divide by std
    all_scaling_vals = {
        "CH02": {"mean": 0.08713, "std": 0.15314},
        "CH04": {"mean": 0.023459, "std": 0.06092},
        "CH05": {"mean": 0.047332, "std": 0.0834376},
        "CH06": {"mean": 0.04413, "std": 0.080274},
        "CH08": {"mean": 234.73875, "std": 6.4855},
        "CH09": {"mean": 245.538996, "std": 10.65327},
        "CH10": {"mean": 252.71758, "std": 11.06057},
        "CH11": {"mean": 274.36105, "std": 21.69784},
        "CH13": {"mean": 270.93, "std": 25.9126},
        "CH15": {"mean": 273.168, "std": 19.9895},
        "CH16": {"mean": 260.5381, "std": 17.2468},
        "FED": {"mean": 0.3536, "std": 2.341},
        "HRAC_COM_FR": {"mean": 0.023433, "std": 0.0305756},
        "LANDMASK": {"mean": 0, "std": 1},
        "sinday": {"mean": 0.0, "std": 0.707},
        "cosday": {"mean": 0.0, "std": 0.707},
        "satzen": {"mean": 42.68, "std": 13.55424},
        "solzen": {"mean": 80.176812, "std": 38.9773},
        "lon": {"mean": -88.4005, "std": 18.07278},
        "lat": {"mean": 30.9988, "std": 10.34049},
        "surface_type": {"mean": 3.127724, "std": 4.07028},
    }

    # for normalizing byte-scaled
    # Basically, map the normal mean and std values to 0-->1.
    # Use the formula found in make_bmps.write_bmps, w/o rounding and 255.999 factor.
    #     (value - vmin) / (vmax - vmin)
    # vmin, vmax, and value should be in the native units. value is either original mean or std.

    norm_scaling_vals = {
        "CH02": {"mean": 0.08713, "std": 0.15314},
        "CH04": {"mean": 0.031279, "std": 0.08123},
        "CH05": {"mean": 0.063109, "std": 0.11125},
        "CH06": {"mean": 0.05884, "std": 0.107032},
        "CH08": {"mean": 0.496268, "std": -2.764493},
        "CH09": {"mean": 0.6071866, "std": -2.524623},
        "CH10": {"mean": 0.65897, "std": -2.36174},
        "CH11": {"mean": 0.648931, "std": -1.294632},
        "CH13": {"mean": 0.622538, "std": -1.26221},
        "CH15": {"mean": 0.639753, "std": -1.307773},
        "CH16": {"mean": 0.641255, "std": -1.57048},
        "FED": {"mean": 0.014144, "std": 0.09364},
        "HRAC_COM_FR": {
            "mean": 0.0976375,
            "std": 0.127398,
        },  # daily 'hires' ltg climatology
        "LANDMASK": {"mean": 0, "std": 1},
    }

    # NB: the HRRR fields are 64x64, but at a different domain than ABI. Thus, they are not technically 1km res.
    hs = {
        "8km": ["HRAC_COM_FR", "CLIMO"],
        "4km": ["GHRSST"],
        "2km": [
            "CH04",
            "CH06",
            "CH07",
            "CH08",
            "CH09",
            "CH10",
            "CH11",
            "CH12",
            "CH13",
            "CH14",
            "CH15",
            "CH16",
            "FED",
            "AFA",
            "FCD",
            "MFA",
            "TOE",
            "BTD1315",
            "BTD1113",
            "u925",
            "v925",
            "u850",
            "v850",
            "u700",
            "v700",
            "u500",
            "v500",
            "mlcape",
            "surface_type",
            "LANDMASK",
        ],
        "1km": ["CH01", "CH03", "CH05", "REF10"],
        "05km": ["CH02"],
    }

    cnn = "unet_2D"  #'unet_[2,3]D' 'convlstm2d'
    data = []  # set to empty if using samplelists or custom netcdfs
    train_dates = val_dates = []  # set to empty if using samplelists (see next section)

    # if train_dates = val_dates = [], then we are using samplelists.

    sat = "goes16"

    train_list = [
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2019/2019??[0,1]?/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2019/2019??2[0-4]/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2020/2020??[0,1]?/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2020/2020??2[0-4]/*tfrec",
    ]
    for patt in train_list:
        ff = glob.glob(patt)
        if len(ff) == 0:
            print("ERROR: Something is wrong with train_list patt:")
            print(patt)
            sys.exit()
    val_list = [
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2019/2019??2[5-9]/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2019/2019??3?/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2020/2020??2[5-9]/*tfrec",
        f"/ships22/grain/probsevere/LC/tfrecs3/{sat}/2020/2020??3?/*tfrec",
    ]
    for patt in val_list:
        ff = glob.glob(patt)
        if len(ff) == 0:
            print("ERROR: Something is wrong with val_list patt:")
            print(patt)
            sys.exit()

    outdir = f"/ships22/grain/probsevere/LC/tests/2019-2020/c02051315_poswt5"

    inputs = []

    # scalar_vars = ['cosday','sinday']
    scalar_vars = []
    if cnn.startswith("unet") or cnn == "convlstm2d":
        ngpu = 2
        batchsize = ngpu * 4
        inputs.append(["CH02"])  # 0.5km
        inputs.append(["CH05"])  # 1km
        inputs.append(["CH13", "CH15"])  # 2km
        # inputs.append(['CLIMO'])
        ps_2km = (480, 480)  # (736,832) #lustre patch size
        # ps_2km = (128,128)
        ps_8km = tuple([_ // 4 for _ in ps_2km])
        ps_4km = tuple([_ // 2 for _ in ps_2km])
        ps_1km = tuple([_ * 2 for _ in ps_2km])
        ps_05km = tuple([_ * 4 for _ in ps_2km])
        input_tuples = []
        for i, inp in enumerate(inputs):  # channels last for tf
            if inp[0] in hs["2km"]:
                input_tuples.append((ps_2km[0], ps_2km[1], len(inp)))
            elif inp[0] in hs["1km"]:
                input_tuples.append((ps_1km[0], ps_1km[1], len(inp)))
            elif inp[0] in hs["05km"]:
                input_tuples.append((ps_05km[0], ps_05km[1], len(inp)))
            elif inp[0] in hs["4km"]:
                input_tuples.append((ps_4km[0], ps_4km[1], len(inp)))
            elif inp[0] in hs["8km"]:
                input_tuples.append((ps_8km[0], ps_8km[1], len(inp)))
            else:
                print("Can't find " + inp + " in hs lists!")
                sys.exit()

        # make the scaling values dict from all_scaling_vals "library"
        scaling_vals = {}  # used when data are in native units
        minmax_scaling_vals = {}  # used in conjunction with byte_scaling_vals
        byte_scaling_vals = {}
        for inp in inputs:
            for ch in inp:
                #    scaling_vals[ch] = all_scaling_vals[ch]
                minmax_scaling_vals[ch] = bsinfo[ch]
        #    byte_scaling_vals[ch] = norm_scaling_vals[ch]

        loss_fcn = "binary_crossentropy"  #'QR' #'binary_crossentropy'

        datagen = "None"  #'DataGeneratorLUSTRE'  #not used, currently

        # We can have up to 2 targets.
        # E.g., P(≥ 1fl in 60 min) AND P(≥ 5fl in 60 min)  Fro this, we need two binarization thresholds (binarize1 and binarize2)
        # E.g., P(≥ 1 fl in 15 min) AND P(≥ 1fl in 60 min) For this, we need two target names (target1 and target2)

        ntargets = 1
        binarize1 = 1  # value such that >= value is set to 1; < value set to 0. This is in *scaled-bitmap space!*
        # Data were scaled 0 to 255. 0.5 fl maps to a value of 1.
        binarize2 = 0

        target1 = "FED_accum_60min_2km"  #'FED_accum_60min_2km' # 'FED_accum_60min_2km' 'MAXFED_5to60'
        target2 = None  #'FED_accum_15min_2km'  #None if not using

        if binarize2 > 0:
            assert binarize1 < binarize2
            assert ntargets == 2
        else:
            if target2:
                assert ntargets == 2
            else:
                assert ntargets == 1
        learning_rate = 0.0001

        class_weights = False  # {0:0.51666592, 1:15.50067251} #computed with binarization threshold = 0.99 and target = 'MAXFED_5to60'
        # {0:0.5400779 , 1:6.73785139} #computed with 400x400 binarization threshold=0.99 and target='MAXFED_5to60'

        # augmentations -- options: 'random_rotation:1', 'random_noise':0.1
        img_aug = {}  #'random_noise':0.1}

        timeind = []  # used for convlstm2d and conv3d

        # architecture
        num_conv_filters = 16
        dilation = 1
        bias_init = None  # np.array([-2.52378297]) #this np.log([pos/neg]) ; see https://www.tensorflow.org/tutorials/structured_data/imbalanced_data

        num_encoding_blocks = 4
        num_conv_per_block = 2
        nfmaps_by_block = [
            num_conv_filters,
            int(num_conv_filters * 2),
            int(num_conv_filters * 4),
            int(num_conv_filters * 8),
            int(num_conv_filters * 16),
        ]  # last element is for middle/bottleneck of unet
        assert len(nfmaps_by_block) == num_encoding_blocks + 1
        num_decoding_blocks = 2

    channels = []
    for inp in inputs:
        channels += inp

    if scalar_vars:
        # assuming input_tuples[-1] is 2km data. This makes the 'scalars' 8km resolution.
        scalar_reduce_res_fac = srrf = 4
        scalar_tuple = (
            len(scalar_vars),
            input_tuples[-1][1] // srrf,
            input_tuples[-1][2] // srrf,
        )
        for sv in scalar_vars:
            scaling_vals[sv] = all_scaling_vals[sv]
            byte_scaling_vals[sv] = norm_scaling_vals[sv]
        # input_tuples.append(scalar_tuple)
        # input_tuples.append((len(scalar_vars),))
    else:
        scalar_reduce_res_fac = 4
        scalar_tuple = ()

    return {
        "class_weights": class_weights,
        "batch_norm": False,
        "bias_init": bias_init,
        "shuffle": shuffle,
        "nframes": nframes,
        "timeind": timeind,
        "scalar_vars": scalar_vars,
        "scalar_tuple": scalar_tuple,
        "ps_8km": ps_8km,
        "ps_2km": ps_2km,
        "ps_1km": ps_1km,
        "ps_05km": ps_05km,
        "channels": channels,
        "l2_param": 0.0,
        "dropout_rate": 0.0,
        "batchsize": batchsize,
        "num_conv_filters": num_conv_filters,
        "num_encoding_blocks": num_encoding_blocks,
        "num_decoding_blocks": num_decoding_blocks,
        "num_conv_per_block": num_conv_per_block,
        "nfmaps_by_block": nfmaps_by_block,
        "filter_width": 3,
        "padding": "same",
        "conv_activation": "leaky_relu",
        "learning_rate": learning_rate,
        "nepoch": 100,
        "nworkers": nworkers,
        "target_time": 60,  # in min
        "es_patience": 3,
        "rlr_factor": 0.1,
        "rlr_patience": 1,
        "rlr_cooldown": 2,
        "stride": 4,
        "cnn": cnn,
        "input_tuples": input_tuples,
        "inputs": inputs,
        "val_dates": val_dates,
        "train_dates": train_dates,
        "train_list": train_list,
        "val_list": val_list,
        "target1": target1,
        "target2": target2,
        "data": data,
        "scaling_vals": scaling_vals,
        "minmax_scaling_vals": minmax_scaling_vals,
        "byte_scaling_vals": byte_scaling_vals,
        "binarize1": binarize1,
        "binarize2": binarize2,
        "ntargets": ntargets,
        "minmax": minmax,
        "scalar_reduce_res_fac": scalar_reduce_res_fac,
        "datagen": datagen,
        "dilation": dilation,
        "ngpu": ngpu,
        "loss_fcn": loss_fcn,
        "optimizer": "adabelief",  # Adam, SGD, adabelief
        "outdir": outdir,
        "img_aug": img_aug,
        "noskips": False,  # if True, won't do the normal skip connections in the U-Net.
        "upsamp": True,  # indicates that we do upsampling to the highest spatial res.
        "spatial_verification": False,
    }


if __name__ == "__main__":
    outdir = sys.argv[1]
    os.makedirs(outdir, exist_ok=True)
    print(tf_config())
    # sys.exit()
    pickle.dump(tf_config(), open(os.path.join(outdir, "model_config.pkl"), "wb"))
