import numpy as np
import netCDF4
import os, sys
import glob as glob

import tensorflow as tf

tf.keras.backend.clear_session()
from tensorflow import keras

import time
from datetime import datetime, timedelta
import matplotlib.pyplot as plt


class DataGeneratorZARR(tf.keras.utils.Sequence):
    "Generates data for Keras"

    def __init__(
        self,
        zarr,
        inds,
        inputs={"input1": {"CH13": (32, 32, 1)}},
        nframes=1,
        permute_input=[(-1, -1)],
        scalars=[],
        target="labels",
        batch_size=64,
        shuffle=True,
        binarize=-1,
    ):
        "Initialization"
        self.inds = inds
        self.inputs = inputs
        self.batch_size = batch_size
        self.scalars = scalars
        self.nframes = nframes
        self.zarr = zarr  # assumes samples is axis 0. This should be an opened zarr
        self.shuffle = shuffle
        self.target = target
        self.permute_input = permute_input
        self.binarize = binarize
        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches per epoch"
        print("nbatches per epoch:", int(np.floor(len(self.inds) / self.batch_size)))
        return int(np.floor(len(self.inds) / self.batch_size))

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.inds))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of samples
        list_inds = [self.inds[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_inds)

        if (
            self.permute_input[0][0] >= 0
        ):  # permute_input is a list of tuples telling which inputs and indexes to permute.
            # For instance, (1,0) means permute the 0th predictor in the first input.
            # this block should only be done for permutation testing. X[input_ind].shape[0] = nsamples
            for pi in self.permute_input:
                input_ind = pi[0]
                pred_ind = pi[1]
                X[input_ind][..., pred_ind] = np.take(
                    X[input_ind][..., pred_ind],
                    indices=np.random.permutation(X[input_ind].shape[0]),
                    axis=0,
                )

        return X, y

    def __data_generation(self, list_inds):
        "Generates data containing batch_size samples"  # X : (n_samples, *dim, n_channels)
        inputs = []

        if self.nframes == 1:
            lastind = [-1]
        else:
            lastind = slice(None)  # all frames

        X1 = None
        for chan in self.inputs["input1"]:
            mean = self.zarr[chan].attrs["mean"]
            std = self.zarr[chan].attrs["std"]
            tmp = (
                self.zarr[chan].get_orthogonal_selection(
                    (list_inds, slice(None), slice(None), lastind)
                )
                - mean
            ) / std
            if X1 is None:
                X1 = tmp + 0.0
            else:
                X1 = np.concatenate((X1, tmp), axis=-1)
        # this is only for 3D conv / 'video'
        #    if(self.nframes > 1): X1 = np.expand_dims(np.transpose(X1,(0,3,1,2)),axis=-1)
        inputs.append(X1)

        X2 = None
        if "input2" in self.inputs:
            for chan in self.inputs["input2"]:
                mean = self.zarr[chan].attrs["mean"]
                std = self.zarr[chan].attrs["std"]
                tmp = (
                    self.zarr[chan].get_orthogonal_selection(
                        (list_inds, slice(None), slice(None), lastind)
                    )
                    - mean
                ) / std
                if X2 is None:
                    X2 = tmp + 0.0
                else:
                    X2 = np.concatenate((X2, tmp), axis=-1)
        #    if(self.nframes > 1 and X2 is not None): X2 = np.expand_dims(np.transpose(X2,(0,3,1,2)),axis=-1)
        if X2 is not None:
            inputs.append(X2)

        X3 = None
        if "input3" in self.inputs:
            for chan in self.inputs["input3"]:
                mean = self.zarr[chan].attrs["mean"]
                std = self.zarr[chan].attrs["std"]
                tmp = (
                    self.zarr[chan].get_orthogonal_selection(
                        (list_inds, slice(None), slice(None), lastind)
                    )
                    - mean
                ) / std
                if X3 is None:
                    X3 = tmp + 0.0
                else:
                    X3 = np.concatenate((X3, tmp), axis=-1)
        if self.nframes > 1 and X3 is not None:
            X3 = np.expand_dims(np.transpose(X3, (0, 3, 1, 2)), axis=-1)
        if X3 is not None:
            inputs.append(X3)

        scalars = None
        for chan in self.scalars:
            mean = self.zarr[chan].attrs["mean"]
            std = self.zarr[chan].attrs["std"]
            tmp = (self.zarr[chan].get_orthogonal_selection(list_inds) - mean) / std
            if scalars is None:
                scalars = tmp + 0.0
            else:
                scalars = np.column_stack((scalars, tmp))
        if scalars is not None:
            inputs.append(scalars)

        # targets
        if self.binarize > 0:  # this doubles for unet flag
            y = self.zarr[self.target].get_orthogonal_selection(
                (list_inds, slice(None), slice(None), [-1])
            )
            y[y < self.binarize] = 0
            y[y >= self.binarize] = 1
            y = np.squeeze(y, axis=-1)  # remove the superfluous last dim
        else:
            if self.target != "labels":  # the default
                y = self.zarr[self.target].get_coordinate_selection(list_inds)
                y[y > 1] = 1  # make "linear" as "supercell"
            else:
                y = np.zeros(len(list_inds))  # dummy list of 'truth'

        return tuple(inputs), y


#############################################################################################################################################
class DataGeneratorSCALE(tf.keras.utils.Sequence):
    "Generates data for Keras"

    def __init__(
        self,
        zarr,
        inds,
        inputs={"input1": {"CH13": (32, 32, 1)}},
        nframes=1,
        permute_input=[(-1, -1)],
        scalars=[],
        target="labels",
        batch_size=64,
        shuffle=True,
    ):
        "Initialization"
        self.inds = inds
        self.inputs = inputs
        self.batch_size = batch_size
        self.scalars = scalars
        self.nframes = nframes
        self.zarr = zarr  # assumes samples is axis 0. This should be an opened zarr
        self.shuffle = shuffle
        self.target = target
        self.permute_input = permute_input
        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches per epoch"
        print("nbatches per epoch:", int(np.floor(len(self.inds) / self.batch_size)))
        return int(np.floor(len(self.inds) / self.batch_size))

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.inds))
        # if self.shuffle == True:
        #  np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of samples
        list_inds = [self.inds[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_inds)

        if (
            self.permute_input[0][0] >= 0
        ):  # permute_input is a list of tuples telling which inputs and indexes to permute.
            # For instance, (1,0) means permute the 0th predictor in the first input.
            # this block should only be done for permutation testing. X[input_ind].shape[0] = nsamples
            for pi in self.permute_input:
                input_ind = pi[0]
                pred_ind = pi[1]
                X[input_ind][..., pred_ind] = np.take(
                    X[input_ind][..., pred_ind],
                    indices=np.random.permutation(X[input_ind].shape[0]),
                    axis=0,
                )

        return X, y

    def __data_generation(self, list_inds):
        "Generates data containing batch_size samples"  # X : (n_samples, *dim, n_channels)
        inputs = []

        if self.nframes == 1:
            lastind = [-1]
        else:
            lastind = slice(None)  # all frames

        X1 = None
        for chan in self.inputs["input1"]:
            tmp = self.zarr[chan].get_orthogonal_selection(
                (list_inds, slice(None), slice(None), lastind)
            )
            tmp[tmp > 0.9] = 0.9
            tmp[tmp < 0] = 0
            tmp = (tmp - 0) / (0.9 - 0)
            if X1 is None:
                X1 = tmp + 0.0
            else:
                X1 = np.concatenate((X1, tmp), axis=-1)
        # this is only for 3D conv / 'video'
        if self.nframes > 1:
            X1 = np.expand_dims(np.transpose(X1, (0, 3, 1, 2)), axis=-1)
        inputs.append(X1)

        X2 = None
        if "input2" in self.inputs:
            for chan in self.inputs["input2"]:
                tmp = self.zarr[chan].get_orthogonal_selection(
                    (list_inds, slice(None), slice(None), lastind)
                )
                tmp[tmp > 0.5] = 0.5
                tmp[tmp < 0] = 0
                tmp = (tmp - 0) / (0.5 - 0)
                if X2 is None:
                    X2 = tmp + 0.0
                else:
                    X2 = np.concatenate((X2, tmp), axis=-1)
        if self.nframes > 1 and X2 is not None:
            X2 = np.expand_dims(np.transpose(X2, (0, 3, 1, 2)), axis=-1)
        if X2 is not None:
            inputs.append(X2)

        X3 = None
        if "input3" in self.inputs:
            for chan in self.inputs["input3"]:
                tmp = self.zarr[chan].get_orthogonal_selection(
                    (list_inds, slice(None), slice(None), lastind)
                )
                tmp[tmp > 293] = 293
                tmp[tmp < 233] = 233
                tmp = (tmp - 233) / (293 - 233)
                if X3 is None:
                    X3 = tmp + 0.0
                else:
                    X3 = np.concatenate((X3, tmp), axis=-1)
        if self.nframes > 1 and X3 is not None:
            X3 = np.expand_dims(np.transpose(X3, (0, 3, 1, 2)), axis=-1)
        if X3 is not None:
            inputs.append(X3)

        scalars = None
        for chan in self.scalars:
            mean = self.zarr[chan].attrs["mean"]
            std = self.zarr[chan].attrs["std"]
            tmp = (self.zarr[chan].get_orthogonal_selection(list_inds) - mean) / std
            if scalars is None:
                scalars = tmp + 0.0
            else:
                scalars = np.column_stack((scalars, tmp))
        if scalars is not None:
            inputs.append(scalars)

        # targets
        y = self.zarr[self.target].get_coordinate_selection(list_inds)

        return tuple(inputs), y


#############################################################################################################################################
class DataGeneratorUNET3D(tf.keras.utils.Sequence):
    "Generates data for Keras"

    def __init__(self, inds, zarr, config):
        "Initialization"
        self.inds = inds
        self.inputs = config["inputs"]
        self.batch_size = config["batchsize"]
        self.timeind = config["timeind"]
        self.scalars = config["scalar_vars"]
        self.nframes = config["nframes"]
        self.zarr = zarr  # assumes samples is axis 0. This should be an opened zarr
        self.shuffle = config["shuffle"]
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scaling_vals = config["scaling_vals"]
        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches per epoch"
        print("nbatches per epoch:", int(np.floor(len(self.inds) / self.batch_size)))
        return int(np.floor(len(self.inds) / self.batch_size))

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.inds))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of samples
        list_inds = [self.inds[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_inds)

        return X, y

    def __data_generation(self, list_inds):
        "Generates data containing batch_size samples"  # X : (n_samples, *dim, n_channels)

        inputs = []

        timeind = self.timeind
        yslice = slice(None)
        xslice = slice(None)
        lastind = slice(None)

        # all inputs should have the same channels. Each input is a time slice.
        X1 = []
        for chan in self.inputs[0]:
            mean = self.scaling_vals[chan]["mean"]
            std = self.scaling_vals[chan]["std"]
            tmp = (
                self.zarr[chan].get_orthogonal_selection(
                    (list_inds, timeind, yslice, xslice, lastind)
                )
                - mean
            ) / std
            X1.append(tmp)

        # not sure how to handle multiple channels per time slice
        inputs.append(X1[0][:, 0, ...])
        inputs.append(X1[0][:, 1, ...])
        inputs.append(X1[0][:, 2, ...])

        # targets
        yslice = slice(None)
        xlices = slice(None)
        y = self.zarr[self.target].get_orthogonal_selection(
            (list_inds, yslice, xslice, [-1])
        )
        y[y < self.binarize] = 0
        y[y >= self.binarize] = 1
        y = np.squeeze(y, axis=-1)  # remove the superfluous last dim

        return tuple(inputs), y


#############################################################################################################################################
class DataGeneratorUNET2D(tf.keras.utils.Sequence):
    "Generates data for Keras"

    def __init__(self, inds, zarr, config, permute_input=[(-1, -1)]):
        "Initialization"
        self.inds = inds
        self.inputs = config["inputs"]
        self.batch_size = config["batchsize"]
        self.scalars = config["scalar_vars"]
        self.nframes = config["nframes"]
        self.zarr = zarr  # assumes samples is axis 0. This should be an opened zarr
        self.shuffle = config["shuffle"]
        self.target = config["target"]
        self.permute_input = permute_input
        self.binarize = config["binarize"]
        self.scaling_vals = config["scaling_vals"]
        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches per epoch"
        print("nbatches per epoch:", int(np.floor(len(self.inds) / self.batch_size)))
        return int(np.floor(len(self.inds) / self.batch_size))

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.inds))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of samples
        list_inds = [self.inds[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_inds)

        if (
            self.permute_input[0][0] >= 0
        ):  # permute_input is a list of tuples telling which inputs and indexes to permute.
            # For instance, (1,0) means permute the 0th predictor in the first input.
            # this block should only be done for permutation testing. X[input_ind].shape[0] = nsamples
            for pi in self.permute_input:
                input_ind = pi[0]
                pred_ind = pi[1]
                X[input_ind][..., pred_ind] = np.take(
                    X[input_ind][..., pred_ind],
                    indices=np.random.permutation(X[input_ind].shape[0]),
                    axis=0,
                )

        return X, y

    def __data_generation(self, list_inds):
        "Generates data containing batch_size samples"  # X : (n_samples, *dim, n_channels)
        inputs = []

        yslice = slice(None)
        xslice = slice(None)
        if self.nframes == 1:
            lastind = [-1]
            secondind = [-1]
            sqa = 1
        else:
            lastind = slice(None)  # all frames
            secondind = slice(None)
            sqa = -1

        for inp in self.inputs:  # a list of lists (non-scalar vars)
            X = None
            for chan in inp:
                mean = self.scaling_vals[chan]["mean"]
                std = self.scaling_vals[chan]["std"]
                tmp = (
                    self.zarr[chan].get_orthogonal_selection(
                        (list_inds, secondind, yslice, xslice, lastind)
                    )
                    - mean
                ) / std
                tmp = np.squeeze(
                    tmp, axis=sqa
                )  # remove time axis; assumes we're using 3D netcdfs

                if X is None:
                    X = tmp + 0.0
                else:
                    X = np.concatenate((X, tmp), axis=-1)
            inputs.append(X)

        scalars = None
        for chan in self.scalars:
            mean = self.scaling_vals[chan]["mean"]
            std = self.scaling_vals[chan]["std"]
            tmp = (self.zarr[chan].get_orthogonal_selection(list_inds) - mean) / std
            if scalars is None:
                scalars = tmp + 0.0
            else:
                scalars = np.column_stack((scalars, tmp))
        if scalars is not None:
            inputs.append(scalars)

        # targets
        yslice = slice(None)
        xlices = slice(None)
        y = self.zarr[self.target].get_orthogonal_selection(
            (list_inds, yslice, xslice, [-1])
        )
        y[y < self.binarize] = 0
        y[y >= self.binarize] = 1
        y = np.squeeze(y, axis=-1).astype(np.int32)  # remove the superfluous last dim

        return tuple(inputs), y
        # return {'input_1':inputs[0],'input_2':inputs[1],'input_3':inputs[2]}, y
        # return (inputs[0],inputs[1],inputs[2]), y


#############################################################################################################################################
class DataGeneratorLUSTRE(tf.keras.utils.Sequence):
    "Characterizes a dataset for Tensorflow"

    def __init__(self, samplelist, config, lats=None, lons=None, satzen=None):
        "Initialization"
        self.samplelist = samplelist  # this is a list of dicts with info on each sample
        self.inputs = config["inputs"]
        self.scaling_vals = config["scaling_vals"]
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scalar_vars = config["scalar_vars"]
        self.scalar_tuple = config["scalar_tuple"]
        self.input_tuples = config["input_tuples"]
        self.srrf = config["scalar_reduce_res_fac"]  # usu. 4
        self.batch_size = config["batchsize"]
        self.shuffle = config["shuffle"]
        self.lats = lats
        self.lons = lons
        self.satzen = satzen

        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches"
        nbpe = int(np.floor(len(self.samplelist) / self.batch_size))
        print("nbatches per epoch:", nbpe)
        return nbpe

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.samplelist))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one BATCH of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # Find list of samples
        # inds = [self.dtlist[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(indexes)

        return X, y

    def __data_generation(self, indexes):
        """
        Generates the data. # X : (n_samples, *dim, n_channels)
        Keys in each sample:
        - 'abidt': <datetime.datetime> for ABI scan
        - 'startX','endX','startY','endY': <int> The coordinate
          endpoints in image space, for *2-km* ABI data.
          With RadC files, this is from the fixed-grid format.
          We are breaking up CONUS scenes, so these will vary,
          with about 6 sets of corner points per ABI dt.
        - 'target_file': <str> Full file name for the accumlated/aggregated
          truth metric. E.g., in /apollo/grain/...../FED_accum_60min/%Y%m%d-%H%M00.netcdf
        - 'CH02'...'CH16': <str> The full filenames for each channel. I think it
          is better to just put the file names here, "precomputed" once
          rather than globbing each epoch.
        """

        inputs = []

        for inp in self.inputs:  # self.inputs is a list of lists (non-scalar vars)
            sampX = None
            for index in indexes:
                X = None
                sample = self.samplelist[index]
                startY = sample["startY"]
                endY = sample["endY"]
                startX = sample["startX"]
                endX = sample["endX"]

                for chan in inp:  # chan is a string and inp is a list
                    abifile = sample[chan]

                    tmp = self._make_abi_data(
                        abifile,
                        ch=chan,
                        startX=startX,
                        endX=endX,
                        startY=startY,
                        endY=endY,
                    )

                    # tmp = np.ma.fix_invalid(tmp) #masks nan/inf; currently unnecessary, as data are properly masked
                    tmp_scaled = np.expand_dims(
                        (tmp - self.scaling_vals[chan]["mean"])
                        / self.scaling_vals[chan]["std"],
                        axis=(0, -1),
                    )  # make samples and channels axis
                    tmp_scaled = tmp_scaled.filled(
                        self.scaling_vals[chan]["mean"]
                    )  # also converts to numpy.ndarray
                    if X is None:
                        X = tmp_scaled + 0.0
                    else:
                        X = np.concatenate((X, tmp_scaled), axis=-1)
                if sampX is None:
                    sampX = X + 0.0
                else:
                    sampX = np.concatenate((sampX, X), axis=0)
            inputs.append(sampX)

        # make scalar input -- note that in a U-Net, these aren't strictly 'scalars'
        if len(self.scalar_vars):
            sy = self.scalar_tuple[-2]
            sx = self.scalar_tuple[-1]  # last input_tuple should be scalars
            sampX = None
            for index in indexes:  # each sample
                X = None
                sample = self.samplelist[index]
                startY = sample["startY"]
                endY = sample["endY"]
                startX = sample["startX"]
                endX = sample["endX"]

                for svar in self.scalar_vars:
                    if svar == "month":
                        tmp = np.full(
                            (sy, sx), sample["abidt"].month
                        )  # make same value everywhere
                    elif svar == "week_of_year":
                        tmp = np.full(
                            (sy, sx), int(sample["abidt"].strftime("%j")) // 7
                        )  # make same value everywhere
                    elif svar == "fortnight_of_year":
                        tmp = np.full(
                            (sy, sx), int(sample["abidt"].strftime("%j")) // 14
                        )  # make same value everywhere
                    elif svar == "cosday":  # should be used with sinday
                        doy = int(sample["abidt"].strftime("%j"))
                        tmp = np.full(
                            (sy, sx), np.cos((2 * np.pi / 365) * doy)
                        )  # make same value everywhere # y = cos((2*pi/365) * x), for cylical value
                        # i.e., Dec31 is like Jan1
                    elif svar == "sinday":  # should be used with cosday
                        doy = int(sample["abidt"].strftime("%j"))
                        tmp = np.full(
                            (sy, sx), np.sin((2 * np.pi / 365) * doy)
                        )  # make same value everywhere # y = sin((2*pi/365) * x), for cylical value
                        # i.e., Dec31 is like Jan1
                    elif svar == "solzen":
                        # assuming we want to reduce resolution. start/end X/Y are for 2-km CONUS FGF
                        lat = self.lats[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                        lon = self.lons[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                        tmp, _ = solar_angles.solar_angles(
                            dt=[sample["abidt"]], lat=lat, lon=lon
                        )
                    elif svar == "lat":
                        # assuming we want to reduce resolution.
                        tmp = self.lats[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    elif svar == "lon":
                        # assuming we want to reduce resolution.
                        tmp = self.lons[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    elif svar == "satzen":
                        # assuming we want to reduce resolution.
                        tmp = self.satzen[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    tmp = np.ma.fix_invalid(tmp)  # masks nan/inf
                    tmp_scaled = np.expand_dims(
                        (tmp - self.scaling_vals[svar]["mean"])
                        / self.scaling_vals[svar]["std"],
                        axis=(0, -1),
                    )  # scale and expand
                    try:
                        tmp_scaled = tmp_scaled.filled(
                            self.scaling_vals[svar]["mean"]
                        )  # fill bad pixels with mean...also converts to numpy.ndarray
                    except AttributeError:
                        pass  # it's already an ndarray, not a MaskedArray
                    if X is None:
                        X = tmp_scaled + 0.0
                    else:
                        X = np.concatenate((X, tmp_scaled), axis=-1)
                if sampX is None:
                    sampX = X + 0.0
                else:
                    sampX = np.concatenate((sampX, X), axis=0)
            inputs.append(sampX)

        # target/truth tensor
        target = None
        for index in indexes:
            sample = self.samplelist[index]
            try:  # truth start/end X/Y, which could be different from start/end X/Y
                startY = sample["tstartY"]
                endY = sample["tendY"]
                startX = sample["tstartX"]
                endX = sample["tendX"]
            except KeyError:
                startY = sample["startY"]
                endY = sample["endY"]
                startX = sample["startX"]
                endX = sample["endX"]

            target_file = sample["target_file"]
            nc = netCDF4.Dataset(target_file, "r")
            # check if we're using 8-km data
            if nc.spatial_resolution == "reduced from 2-km to 8-km":
                y = np.expand_dims(
                    nc.variables[self.target][
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ],
                    axis=0,
                )
                increase_res = True
            else:  # else assume 2-km target data
                y = np.expand_dims(
                    nc.variables[self.target][startY:endY, startX:endX], axis=0
                )
                increase_res = False
            y = y.filled(
                0
            )  # masked values become 0 and y becomes np.ndarray instead of np.ma.core.MaskedArray (required by pytorch)
            y[y < self.binarize] = 0
            y[y >= self.binarize] = 1
            if increase_res:
                y = y.repeat(4, axis=1).repeat(
                    4, axis=2
                )  # artificially restore to 2-km res. 0th is channels dim.
            nc.close()

            if target is None:
                target = y + 0.0
            else:
                target = np.concatenate((target, y), axis=0)

        return tuple(inputs), target

    def _make_abi_data(
        self, abifile, ch=None, startY=0, endY=1500, startX=0, endX=2500
    ):
        "Return the possibly cropped reflectance or BT"
        nc = netCDF4.Dataset(abifile, "r")
        if ch is None:  # get chan from file name
            basefile = os.path.basename(abifile)
            ch = "CH" + basefile.split("_")[1].split("-")[-1].split("C")[-1]

        if ch == "surface_type":
            result = nc.variables["pixel_surface_type"][startY:endY, startX:endX]
        elif ch == "CH02":
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 4 : endY * 4, startX * 4 : endX * 4]
                * nc.variables["kappa0"][:]
            )
        elif ch in ["CH01", "CH03", "CH05"]:
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 2 : endY * 2, startX * 2 : endX * 2]
                * nc.variables["kappa0"][:]
            )
        elif ch in ["CH04", "CH06"]:
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY:endY, startX:endX]
                * nc.variables["kappa0"][:]
            )
        else:  # IR channels
            # convert IR radiance to BT
            radiance = nc.variables["Rad"][startY:endY, startX:endX]
            result = (
                nc.variables["planck_fk2"][:]
                / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                - nc.variables["planck_bc1"][:]
            ) / nc.variables["planck_bc2"][:]
        nc.close()
        return result


#############################################################################################################################################
class DataGeneratorLUSTREPermutation(tf.keras.utils.Sequence):
    "Characterizes a dataset for Tensorflow"

    def __init__(
        self, samplelist, config, lats=None, lons=None, satzen=None, permute_input=[]
    ):
        "Initialization"
        self.samplelist = samplelist  # this is a list of dicts with info on each sample
        self.nsamples = len(samplelist)
        self.inputs = config["inputs"]
        self.scaling_vals = config["scaling_vals"]
        self.target = config["target"]
        self.binarize = config["binarize"]
        self.scalar_vars = config["scalar_vars"]
        self.scalar_tuple = config["scalar_tuple"]
        self.input_tuples = config["input_tuples"]
        self.srrf = config["scalar_reduce_res_fac"]  # usu. 4
        self.batch_size = config["batchsize"]
        self.shuffle = False  # config['shuffle'] No shuffling, as we're not training!
        self.channels = config["channels"]
        self.lats = lats
        self.lons = lons
        self.satzen = satzen
        self.permute_input = permute_input

        self.on_epoch_end()

    def __len__(self):
        "Denotes the number of batches"
        nbpe = int(np.floor(len(self.samplelist) / self.batch_size))
        print("nbatches per epoch:", nbpe)
        return nbpe

    def on_epoch_end(self):
        "Updates indexes after each epoch"
        self.indexes = np.arange(len(self.samplelist))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __getitem__(self, index):
        "Generate one BATCH of data"
        # Generate indexes of the batch
        indexes = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]

        # indexes gets transformed into a list of dicts, with each channel having a potentially different int index
        new_indexes = []
        for i in range(len(indexes)):  # indexes should generally be len == 1
            new_index = {"target": indexes[i]}  # don't permute the target!
            for chan in self.channels:
                if chan in self.permute_input:
                    new_index[chan] = int(
                        np.random.random() * (self.nsamples - 1)
                    )  # get random sample
                else:
                    new_index[chan] = indexes[
                        i
                    ]  # set channel to normal/unpermuted index
            new_indexes.append(new_index)

        # Generate data
        X, y = self.__data_generation(new_indexes)

        return X, y

    def __data_generation(self, indexes):
        """
        Generates the data. # X : (n_samples, *dim, n_channels)
        Keys in each sample:
        - 'abidt': <datetime.datetime> for ABI scan
        - 'startX','endX','startY','endY': <int> The coordinate
          endpoints in image space, for *2-km* ABI data.
          With RadC files, this is from the fixed-grid format.
          We are breaking up CONUS scenes, so these will vary,
          with about 6 sets of corner points per ABI dt.
        - 'target_file': <str> Full file name for the accumlated/aggregated
          truth metric. E.g., in /apollo/grain/...../FED_accum_60min/%Y%m%d-%H%M00.netcdf
        - 'CH02'...'CH16': <str> The full filenames for each channel. I think it
          is better to just put the file names here, "precomputed" once
          rather than globbing each epoch.
        """

        inputs = []

        for inp in self.inputs:  # self.inputs is a list of lists (non-scalar vars)
            sampX = None
            for (
                index
            ) in (
                indexes
            ):  # remember, indexes is a list of dicts e.g., indexes = [{'CH02':5, 'CH05':5, 'CH13':22}]
                X = None

                for chan in inp:  # chan is a string and inp is a list
                    sample = self.samplelist[
                        index[chan]
                    ]  # a potentially different sample for each channel
                    startY = sample["startY"]
                    endY = sample["endY"]
                    startX = sample["startX"]
                    endX = sample["endX"]

                    abifile = sample[chan]

                    tmp = self._make_abi_data(
                        abifile,
                        ch=chan[2:],
                        startX=startX,
                        endX=endX,
                        startY=startY,
                        endY=endY,
                    )

                    # tmp = np.ma.fix_invalid(tmp) #masks nan/inf; currently unnecessary, as data are properly masked
                    tmp_scaled = np.expand_dims(
                        (tmp - self.scaling_vals[chan]["mean"])
                        / self.scaling_vals[chan]["std"],
                        axis=(0, -1),
                    )  # make samples and channels axis
                    tmp_scaled = tmp_scaled.filled(
                        self.scaling_vals[chan]["mean"]
                    )  # also converts to numpy.ndarray
                    if X is None:
                        X = tmp_scaled + 0.0
                    else:
                        X = np.concatenate((X, tmp_scaled), axis=-1)
                if sampX is None:
                    sampX = X + 0.0
                else:
                    sampX = np.concatenate((sampX, X), axis=0)
            inputs.append(sampX)

        # make scalar input -- note that in a U-Net, these aren't strictly 'scalars'
        if len(self.scalar_vars):
            sy = self.scalar_tuple[-2]
            sx = self.scalar_tuple[-1]  # last input_tuple should be scalars
            sampX = None
            for (
                index
            ) in (
                indexes
            ):  # each sample FIXME this may need changing if we use "scalars"
                X = None
                sample = self.samplelist[index]
                startY = sample["startY"]
                endY = sample["endY"]
                startX = sample["startX"]
                endX = sample["endX"]

                for svar in self.scalar_vars:
                    if svar == "month":
                        tmp = np.full(
                            (sy, sx), sample["abidt"].month
                        )  # make same value everywhere
                    elif svar == "week_of_year":
                        tmp = np.full(
                            (sy, sx), int(sample["abidt"].strftime("%j")) // 7
                        )  # make same value everywhere
                    elif svar == "fortnight_of_year":
                        tmp = np.full(
                            (sy, sx), int(sample["abidt"].strftime("%j")) // 14
                        )  # make same value everywhere
                    elif svar == "cosday":  # should be used with sinday
                        doy = int(sample["abidt"].strftime("%j"))
                        tmp = np.full(
                            (sy, sx), np.cos((2 * np.pi / 365) * doy)
                        )  # make same value everywhere # y = cos((2*pi/365) * x), for cylical value
                        # i.e., Dec31 is like Jan1
                    elif svar == "sinday":  # should be used with cosday
                        doy = int(sample["abidt"].strftime("%j"))
                        tmp = np.full(
                            (sy, sx), np.sin((2 * np.pi / 365) * doy)
                        )  # make same value everywhere # y = sin((2*pi/365) * x), for cylical value
                        # i.e., Dec31 is like Jan1
                    elif svar == "solzen":
                        # assuming we want to reduce resolution. start/end X/Y are for 2-km CONUS FGF
                        lat = self.lats[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                        lon = self.lons[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                        tmp, _ = solar_angles.solar_angles(
                            dt=[sample["abidt"]], lat=lat, lon=lon
                        )
                    elif svar == "lat":
                        # assuming we want to reduce resolution.
                        tmp = self.lats[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    elif svar == "lon":
                        # assuming we want to reduce resolution.
                        tmp = self.lons[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    elif svar == "satzen":
                        # assuming we want to reduce resolution.
                        tmp = self.satzen[
                            startY // self.srrf : endY // self.srrf,
                            startX // self.srrf : endX // self.srrf,
                        ]
                    tmp = np.ma.fix_invalid(tmp)  # masks nan/inf
                    tmp_scaled = np.expand_dims(
                        (tmp - self.scaling_vals[svar]["mean"])
                        / self.scaling_vals[svar]["std"],
                        axis=(0, -1),
                    )  # scale and expand
                    try:
                        tmp_scaled = tmp_scaled.filled(
                            self.scaling_vals[svar]["mean"]
                        )  # fill bad pixels with mean...also converts to numpy.ndarray
                    except AttributeError:
                        pass  # it's already an ndarray, not a MaskedArray
                    if X is None:
                        X = tmp_scaled + 0.0
                    else:
                        X = np.concatenate((X, tmp_scaled), axis=-1)
                if sampX is None:
                    sampX = X + 0.0
                else:
                    sampX = np.concatenate((sampX, X), axis=0)
            inputs.append(sampX)

        # target/truth tensor
        target = None
        for index in indexes:
            sample = self.samplelist[index["target"]]
            try:  # truth start/end X/Y, which could be different from start/end X/Y
                startY = sample["tstartY"]
                endY = sample["tendY"]
                startX = sample["tstartX"]
                endX = sample["tendX"]
            except KeyError:
                startY = sample["startY"]
                endY = sample["endY"]
                startX = sample["startX"]
                endX = sample["endX"]

            target_file = sample["target_file"]
            nc = netCDF4.Dataset(target_file, "r")
            # check if we're using 8-km data
            if nc.spatial_resolution == "reduced from 2-km to 8-km":
                y = np.expand_dims(
                    nc.variables[self.target][
                        startY // self.srrf : endY // self.srrf,
                        startX // self.srrf : endX // self.srrf,
                    ],
                    axis=0,
                )
                increase_res = True
            else:  # else assume 2-km target data
                y = np.expand_dims(
                    nc.variables[self.target][startY:endY, startX:endX], axis=0
                )
                increase_res = False
            y = y.filled(
                0
            )  # masked values become 0 and y becomes np.ndarray instead of np.ma.core.MaskedArray (required by pytorch)
            y[y < self.binarize] = 0
            y[y >= self.binarize] = 1
            if increase_res:
                y = y.repeat(4, axis=1).repeat(
                    4, axis=2
                )  # artificially restore to 2-km res. 0th is channels dim.
            nc.close()

            if target is None:
                target = y + 0.0
            else:
                target = np.concatenate((target, y), axis=0)

        return tuple(inputs), target

    def _make_abi_data(
        self, abifile, ch=None, startY=0, endY=1500, startX=0, endX=2500
    ):
        "Return the possibly cropped reflectance or BT"
        nc = netCDF4.Dataset(abifile, "r")
        if ch is None:  # get chan from file name
            basefile = os.path.basename(abifile)
            ch = basefile.split("_")[1].split("-")[-1].split("C")[-1]

        if ch == "02":
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 4 : endY * 4, startX * 4 : endX * 4]
                * nc.variables["kappa0"][:]
            )
        elif ch in ["01", "03", "05"]:
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY * 2 : endY * 2, startX * 2 : endX * 2]
                * nc.variables["kappa0"][:]
            )
        elif ch in ["04", "06"]:
            # convert vis radiance to reflectance
            result = (
                nc.variables["Rad"][startY:endY, startX:endX]
                * nc.variables["kappa0"][:]
            )
        else:  # IR channels
            # convert IR radiance to BT
            radiance = nc.variables["Rad"][startY:endY, startX:endX]
            result = (
                nc.variables["planck_fk2"][:]
                / (np.log((nc.variables["planck_fk1"][:] / radiance) + 1))
                - nc.variables["planck_bc1"][:]
            ) / nc.variables["planck_bc2"][:]
        nc.close()
        return result


#############################################################################################################################################
class DataGeneratorPredict(tf.keras.utils.Sequence):
    "Inference only for U-nets"

    def __init__(self, input_list, config_dict, corners, scalars=None, batch_size=128):
        "Initialization"
        self.input_list = input_list  # input_list contains lists of input grids (or sequences of grids)
        # For instance, if your model has two inputs with two channels
        # in input 1 and one channel in input 2, it would look
        # like this [[(ny1,nx1),(ny1,nx1)], [(ny2,nx2,]]
        self.input_tuples = config_dict[
            "input_tuples"
        ]  # a list of containing the shapes of each input
        # self.batch_size = batch_size
        self.ninputs = len(input_list)  # this is only for 2D+ inputs! (not scalars)
        self.svals = config_dict["scaling_values"]
        self.config = config_dict
        self.scalars = scalars
        self.input_corners = corners  # a list of lists of tuples for each NW corner point to feed to the model. Should be len(input_list)

        self.nsamples = len(self.input_corners[0])
        self.batch_size = batch_size  # len(self.input_corners[0])
        self.indexes = np.arange(self.nsamples, dtype=int)

    def __len__(self):
        "Denotes the number of batches"
        print("nsamples:", self.nsamples, "batch size:", self.batch_size)
        print("nbatches:", int(np.ceil(self.nsamples / self.batch_size)))
        return int(np.ceil(self.nsamples / self.batch_size))

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        inds = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        f = inds[0]
        l = inds[-1] + 1

        X = self.__data_generation(f, l)
        return X

    def __data_generation(self, f, l):
        # Generate data
        X = []
        for i in range(len(self.input_tuples)):
            list_in = list(
                self.input_tuples[i]
            )  # we need to prepend the batch_size to each tuple
            list_in.insert(0, self.batch_size)
            X.append(np.empty(tuple(list_in)))

        # For each 2D+ channel in each input, add the patch of data to the preallocated X
        for i in range(self.ninputs):  # 2D+ inputs only! (not scalars)
            for ch in range(len(self.input_list[i])):
                chan = self.config["inputs"]["input" + str(i + 1)][
                    ch
                ]  # this is the actual string of the channel
                ny = self.input_tuples[i][0]
                nx = self.input_tuples[i][1]
                p = 0  # p == "patch
                for c in self.input_corners[i][
                    f:l
                ]:  # corners go across, then down the grid
                    X[i][p, ..., ch] = (
                        self.input_list[i][ch][..., c[0] : c[0] + ny, c[1] : c[1] + nx]
                        - self.svals.loc[chan, "mean"]
                    ) / self.svals.loc[chan, "std"]
                    p += 1

        # scalar vars
        if self.scalars is not None:
            X[-1] = self.scalars[f:l]

        return X


###############################################################################################################################
class DataGeneratorPredictPatches(tf.keras.utils.Sequence):
    "Inference only for small patch-based model"

    def __init__(self, input_list, input_tuples, centers, scalars=None, batch_size=128):
        "Initialization"
        self.input_list = input_list  # input_list contains lists of input grids (or sequences of grids)
        # For instance, if your model has two inputs with two channels
        # in input 1 and one channel in input 2, it would look
        # like this [[(ny1,nx1),(ny1,nx1)], [(ny2,nx2,]]
        self.input_tuples = (
            input_tuples  # a list of containing the shapes of each input
        )
        self.centers = centers  # centers is a list containing lists of tuples for each input in input_list (should be len(input_list))
        self.nsamples = len(centers[0])
        self.batch_size = batch_size
        self.ninputs = len(input_list)  # this is only for 2D+ inputs! (not scalars)
        self.indexes = np.arange(self.nsamples, dtype=int)
        self.scalars = scalars
        self.hs = (
            []
        )  # hs is a list containing ints of half sizes, one for each input in input_list (should be len(input_list))
        for inp in self.input_tuples:
            if len(inp) > 1:  # i.e., not scalars
                self.hs.append(int(inp[-2] / 2))

    def __len__(self):
        "Denotes the number of batches"
        print("nbatches:", int(np.floor(self.nsamples / self.batch_size)))
        return int(np.floor(self.nsamples / self.batch_size))

    def __getitem__(self, index):
        "Generate one batch of data"
        # Generate indexes of the batch
        inds = self.indexes[index * self.batch_size : (index + 1) * self.batch_size]
        f = inds[0]
        l = inds[-1] + 1

        X = self.__data_generation(f, l)
        return X

    def __data_generation(self, f, l):
        # Generate data
        hs = self.hs
        X = []
        for i in range(len(self.input_tuples)):
            list_in = list(
                self.input_tuples[i]
            )  # we need to prepend the batch_size to each tuple
            list_in.insert(0, self.batch_size)
            X.append(np.empty(tuple(list_in)))

        # For each 2D+ channel in each input, add the patch of data to the preallocated X
        for i in range(self.ninputs):  # 2D+ inputs only! (not scalars)
            for ch in range(len(self.input_list[i])):
                p = 0  # p == "patch
                for c in self.centers[i][f:l]:
                    X[i][p, ..., ch] = self.input_list[i][ch][
                        ..., c[0] - hs[i] : c[0] + hs[i], c[1] - hs[i] : c[1] + hs[i]
                    ]
                    p += 1

        # scalar vars
        if self.scalars is not None:
            X[-1] = self.scalars[f:l]

        return X
