import numpy as np
import logging
from logging.config import dictConfig
from lightningcast.logging_def import *
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.models import Model
import tensorflow.keras.layers
from tensorflow.keras.layers import (
    Dense,
    Activation,
    Conv3D,
    Conv2D,
    Input,
    AveragePooling2D,
    MaxPooling2D,
    Flatten,
    LeakyReLU,
    Dropout,
    LSTM,
    TimeDistributed,
    ConvLSTM2D,
)
from tensorflow.keras.layers import (
    concatenate,
    SpatialDropout2D,
    BatchNormalization,
    Conv3DTranspose,
    Conv2DTranspose,
    UpSampling2D,
    UpSampling3D,
    MaxPooling3D,
    Reshape,
    Rescaling,
)
import tensorflow.keras.optimizers
from tensorflow.keras.regularizers import l2
from lightningcast import keras_metrics
from lightningcast import losses

# from keras_radam import RAdam
import sys

# from focal_loss import BinaryFocalLoss
import tensorflow_addons as tfa

##############################################################################################
def get_metrics():

    csi05 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.05,
        name="csi05",
    )
    csi10 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.1,
        name="csi10",
    )
    csi15 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.15,
        name="csi15",
    )
    csi20 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.2,
        name="csi20",
    )
    csi25 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.25,
        name="csi25",
    )
    csi30 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.3,
        name="csi30",
    )
    csi35 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.35,
        name="csi35",
    )
    csi40 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.4,
        name="csi40",
    )
    csi45 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.45,
        name="csi45",
    )
    csi50 = losses.csi(
        use_as_loss_function=False,
        use_soft_discretization=False,
        hard_discretization_threshold=0.5,
        name="csi50",
    )

    metrics = [
        "accuracy",
        tf.keras.metrics.AUC(name="aupr", curve="PR"),
        tf.keras.metrics.MeanSquaredError(name="brier_score"),
        csi05,
        csi10,
        csi15,
        csi20,
        csi25,
        csi30,
        csi35,
        csi40,
        csi45,
        csi50,
    ]
    return metrics


##############################################################################################
def unet_3D(config):
    # build the CNN
    logging.info("Building the CNN...")
    input_tuples = config["input_tuples"]
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    batch_norm = config["batch_norm"]
    dor = config["dropout_rate"]
    dilation = config["dilation"]

    input_1 = Input((None, None, input_tuples[0][-1]))

    inputs = [input_1]
    concat_inputs = [input_1]

    if False and len(input_tuples) > 1:  # we have multiple resolutions
        upsamp_res = int(input_tuples[0][-2] / input_tuples[1][-2])
        input_2 = Input(
            shape=(None, None, input_tuples[1][-1])
        )  # Input(shape=input_tuples[1]) #Input(shape=(None,None,input_tuples[1][-1]))
        input_2_interp = UpSampling2D(
            size=(upsamp_res, upsamp_res), data_format="channels_last"
        )(input_2)
        inputs.append(input_2)
        concat_inputs.append(input_2_interp)
        if len(input_tuples) > 2:
            upsamp_res = int(input_tuples[0][-2] / input_tuples[2][-2])
            input_3 = Input(
                shape=(None, None, input_tuples[2][-1])
            )  # Input(shape=input_tuples[2]) #Input(shape=(None,None,input_tuples[2][-1]))
            input_3_interp = UpSampling2D(
                size=(upsamp_res, upsamp_res), data_format="channels_last"
            )(input_3)
            inputs.append(input_3)
            concat_inputs.append(input_3_interp)
            if (
                len(input_tuples) > 3
            ):  # ok, this assumes input1 is 0.5km, input2 is 1km, input3 is 2km, and input4 is SST, '4km'
                upsamp_res = int(input_tuples[0][-2] / input_tuples[3][-2])
                input_4 = Input(shape=input_tuples[3])
                input_4_interp = UpSampling2D(
                    size=(upsamp_res, upsamp_res), data_format="channels_last"
                )(input_4)
                inputs.append(input_4)
                concat_inputs.append(input_4_interp)

        final_input = concatenate(concat_inputs)
    else:
        final_input = input_1

    # input1 is 20 min ago, input2 is 10 min ago, input3 is 0 min ago....just 2km for now!
    input_2 = Input((None, None, input_tuples[1][-1]))
    input_3 = Input((None, None, input_tuples[2][-1]))
    inputs = [input_1, input_2, input_3]
    final_input = concatenate([input_1, input_2, input_3])

    def lag_branch(inp, num_conv_filters, filter_width, dilation_rate=dilation):
        for bb, factor in enumerate([1, 1.5, 2, 3]):  # num blocks = len of list here
            conv_in1 = inp if (bb == 0) else conv
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv_in1)
            conv = Activation(conv_activation)(conv)
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv)
            conv = Activation(conv_activation)(conv)
            conv = MaxPooling2D(pool_size=(2, 2))(conv)
        return conv

    # t - 2 branch
    t2 = lag_branch(input_1, num_conv_filters, filter_width, dilation_rate=dilation)

    # t - 1 branch
    t1 = lag_branch(input_2, num_conv_filters, filter_width, dilation_rate=dilation)

    # encoding branch3
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(input_3)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv1)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    if dor > 0:
        pool1 = Dropout(dor / 2.0)(pool1)

    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(pool1)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv2)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    if dor > 0:
        pool2 = Dropout(dor)(pool2)

    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(pool2)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv3)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    if dor > 0:
        pool3 = Dropout(dor)(pool3)

    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(pool3)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv4)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(conv4)
    if dor > 0:
        pool4 = Dropout(dor)(pool4)

    # make the time dim and concatenate along that dim
    t1 = tf.expand_dims(t1, 1)
    t2 = tf.expand_dims(t2, 1)
    pool4 = tf.expand_dims(pool4, 1)

    concat = concatenate([t1, t2, pool4], axis=1)

    conv = Conv3D(
        filters=num_conv_filters * 3, kernel_size=filter_width, padding="same"
    )(concat)
    conv = Conv3D(
        filters=num_conv_filters * 3, kernel_size=filter_width, padding="same"
    )(conv)
    conv = Conv3D(
        filters=num_conv_filters * 3, kernel_size=filter_width, padding="same"
    )(conv)
    conv = Conv3D(
        filters=num_conv_filters * 3, kernel_size=filter_width, padding="same"
    )(conv)
    conv = MaxPooling3D(pool_size=(3, 1, 1))(conv)

    conv = tf.keras.backend.squeeze(conv, axis=1)

    up = Conv2DTranspose(num_conv_filters * 3, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv4], axis=-1)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 2, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv3], axis=-1)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(
        num_conv_filters * 1.5, filter_width, strides=2, padding="same"
    )(
        conv
    )  # no activation here?
    merge = concatenate([up, conv2], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 1, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv1], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    # set initial bias initialization
    output_bias = (
        tf.keras.initializers.Constant(config["bias_init"])
        if (config["bias_init"] is not None)
        else "zeros"
    )

    conv = Conv2D(1, (1, 1), bias_initializer=output_bias)(conv)
    conv = Activation("sigmoid")(conv)
    if "CH02" in config["inputs"][0]:  # reduce "500m" to "2km"
        conv = MaxPooling2D(pool_size=(4, 4))(conv)
    elif "CH05" in config["inputs"][0]:  # reduce "1km" to "2km"
        conv = MaxPooling2D(pool_size=(2, 2))(conv)

    # remove superfluous dims
    conv = tf.keras.backend.squeeze(conv, -1)

    model = Model(inputs=inputs, outputs=conv)

    loss_fcn = config["loss_fcn"]

    opt = getattr(tensorflow.keras.optimizers, config["optimizer"])

    model.compile(optimizer=opt(lr=learning_rate), loss=loss_fcn)

    print(model.summary())

    return model


##############################################################################################
def convlstm2d(config):
    # build the CNN
    logging.info("Building the CNN...")
    input_tuples = config["input_tuples"]
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    batch_norm = config["batch_norm"]
    dor = config["dropout_rate"]
    dilation = config["dilation"]

    input_1 = Input((None, None, input_tuples[0][-1]))

    inputs = [input_1]
    concat_inputs = [input_1]

    if False and len(input_tuples) > 1:  # we have multiple resolutions
        upsamp_res = int(input_tuples[0][-2] / input_tuples[1][-2])
        input_2 = Input(
            shape=(None, None, input_tuples[1][-1])
        )  # Input(shape=input_tuples[1]) #Input(shape=(None,None,input_tuples[1][-1]))
        input_2_interp = UpSampling2D(
            size=(upsamp_res, upsamp_res), data_format="channels_last"
        )(input_2)
        inputs.append(input_2)
        concat_inputs.append(input_2_interp)
        if len(input_tuples) > 2:
            upsamp_res = int(input_tuples[0][-2] / input_tuples[2][-2])
            input_3 = Input(
                shape=(None, None, input_tuples[2][-1])
            )  # Input(shape=input_tuples[2]) #Input(shape=(None,None,input_tuples[2][-1]))
            input_3_interp = UpSampling2D(
                size=(upsamp_res, upsamp_res), data_format="channels_last"
            )(input_3)
            inputs.append(input_3)
            concat_inputs.append(input_3_interp)
            if (
                len(input_tuples) > 3
            ):  # ok, this assumes input1 is 0.5km, input2 is 1km, input3 is 2km, and input4 is SST, '4km'
                upsamp_res = int(input_tuples[0][-2] / input_tuples[3][-2])
                input_4 = Input(shape=input_tuples[3])
                input_4_interp = UpSampling2D(
                    size=(upsamp_res, upsamp_res), data_format="channels_last"
                )(input_4)
                inputs.append(input_4)
                concat_inputs.append(input_4_interp)

        final_input = concatenate(concat_inputs)
    else:
        final_input = input_1

    # input1 is 20 min ago, input2 is 10 min ago, input3 is 0 min ago....just 2km for now!
    input_2 = Input((None, None, input_tuples[1][-1]))
    input_3 = Input((None, None, input_tuples[2][-1]))
    inputs = [input_1, input_2, input_3]
    final_input = concatenate([input_1, input_2, input_3])

    def lag_branch(inp, num_conv_filters, filter_width, dilation_rate=dilation):
        for bb, factor in enumerate([1, 1.5, 2, 3]):  # num blocks = len of list here
            conv_in1 = inp if (bb == 0) else conv
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv_in1)
            conv = Activation(conv_activation)(conv)
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv)
            conv = Activation(conv_activation)(conv)
            conv = MaxPooling2D(pool_size=(2, 2))(conv)
        return conv

    # t - 2 branch
    t2 = lag_branch(input_1, num_conv_filters, filter_width, dilation_rate=dilation)

    # t - 1 branch
    t1 = lag_branch(input_2, num_conv_filters, filter_width, dilation_rate=dilation)

    # encoding branch3
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(input_3)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv1)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    if dor > 0:
        pool1 = Dropout(dor / 2.0)(pool1)

    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(pool1)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv2)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    if dor > 0:
        pool2 = Dropout(dor)(pool2)

    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(pool2)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv3)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    if dor > 0:
        pool3 = Dropout(dor)(pool3)

    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(pool3)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv4)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(conv4)
    if dor > 0:
        pool4 = Dropout(dor)(pool4)

    # make the time dim and concatenate along that dim
    t1 = tf.expand_dims(t1, 1)
    t2 = tf.expand_dims(t2, 1)
    pool4 = tf.expand_dims(pool4, 1)

    concat = concatenate([t1, t2, pool4], axis=1)

    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(concat)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(conv)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(conv)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=False,
    )(conv)

    up = Conv2DTranspose(num_conv_filters * 3, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv4], axis=-1)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 2, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv3], axis=-1)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(
        num_conv_filters * 1.5, filter_width, strides=2, padding="same"
    )(
        conv
    )  # no activation here?
    merge = concatenate([up, conv2], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 1, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv1], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    # set initial bias initialization
    output_bias = (
        tf.keras.initializers.Constant(config["bias_init"])
        if (config["bias_init"] is not None)
        else "zeros"
    )

    conv = Conv2D(1, (1, 1), bias_initializer=output_bias)(conv)
    conv = Activation("sigmoid")(conv)
    if "CH02" in config["inputs"][0]:  # reduce "500m" to "2km"
        conv = MaxPooling2D(pool_size=(4, 4))(conv)
    elif "CH05" in config["inputs"][0]:  # reduce "1km" to "2km"
        conv = MaxPooling2D(pool_size=(2, 2))(conv)

    # remove superfluous dims
    conv = tf.keras.backend.squeeze(conv, -1)

    model = Model(inputs=inputs, outputs=conv)

    loss_fcn = config["loss_fcn"]
    opt = getattr(tensorflow.keras.optimizers, config["optimizer"])

    model.compile(optimizer=opt(lr=learning_rate), loss=loss_fcn)

    print(model.summary())

    return model


##################################################################################################################################
def unet_2D(config):

    input_tuples = config["input_tuples"]
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    batch_norm = config["batch_norm"]
    dor = config["dropout_rate"]
    dilation = config["dilation"]
    padding = config["padding"]
    upsamp = config["upsamp"]
    noskips = config["noskips"]
    ntargets = config["ntargets"]
    loss_fcn = config[
        "loss_fcn"
    ]  # tfa.losses.SigmoidFocalCrossEntropy() #BinaryFocalLoss(gamma=2) #config['loss_fcn']

    ki = "glorot_uniform"  # he_normal; glorot_uniform (default)

    incter = 0
    input_1 = Input((None, None, input_tuples[0][-1]), name=f"input_{incter}")
    incter += 1
    inputs = [input_1]
    # input_1  = Rescaling(1./255)(input_1)
    concat_inputs = [input_1]

    if upsamp and len(input_tuples) > 1:  # we have multiple resolutions
        upsamp_res = int(input_tuples[0][-2] / input_tuples[1][-2])
        input_2 = Input(shape=(None, None, input_tuples[1][-1]), name=f"input_{incter}")
        incter += 1
        inputs.append(input_2)
        # input_2  = Rescaling(1./255)(input_2)
        input_2_interp = UpSampling2D(
            size=(upsamp_res, upsamp_res), data_format="channels_last"
        )(input_2)
        concat_inputs.append(input_2_interp)
        if len(input_tuples) > 2:
            upsamp_res = int(input_tuples[0][-2] / input_tuples[2][-2])
            input_3 = Input(
                shape=(None, None, input_tuples[2][-1]), name=f"input_{incter}"
            )
            incter += 1
            inputs.append(input_3)
            # input_3  = Rescaling(1./255)(input_3)
            input_3_interp = UpSampling2D(
                size=(upsamp_res, upsamp_res), data_format="channels_last"
            )(input_3)
            concat_inputs.append(input_3_interp)
            if len(input_tuples) > 3:
                upsamp_res = int(
                    input_tuples[2][-2] / input_tuples[3][-2]
                )  # Going from 8-km to 2-km!
                input_4 = Input(
                    shape=(None, None, input_tuples[3][-1]), name=f"input_{incter}"
                )
                incter += 1
                inputs.append(input_4)
                # input_4  = Rescaling(1./255)(input_4)
                input_4_interp = UpSampling2D(
                    size=(upsamp_res, upsamp_res), data_format="channels_last"
                )(input_4)
                # concat_inputs.append(input_4_interp)

        final_input = concatenate(concat_inputs)
    else:
        final_input = input_1

    num_encoding_blocks = config["num_encoding_blocks"]
    num_conv_per_block = config["num_conv_per_block"]
    nfmaps_by_block = config["nfmaps_by_block"]
    num_decoding_blocks = config["num_decoding_blocks"]

    # encoding branch
    to_concat = []
    for ii in range(num_encoding_blocks):
        nfmaps = nfmaps_by_block[ii]

        # if(not(upsamp)): #add X-km data here (i==0 is 0.5km res., i==1 is 1km res., i==2 is 2km res.)
        if ii == 2 and len(input_tuples) > 3:
            # nchan = input_tuples[ii][-1]
            # new_input = Input(shape=(None,None,nchan), name=f"input_{incter}"); incter+=1
            # inputs.append(new_input)
            conv = concatenate([conv, input_4_interp], axis=-1)

        for jj in range(num_conv_per_block):
            first_block_input = final_input if (ii == 0 and jj == 0) else conv
            conv = Conv2D(
                nfmaps,
                filter_width,
                padding=padding,
                dilation_rate=dilation,
                kernel_initializer=ki,
            )(first_block_input)
            if batch_norm:
                conv = BatchNormalization()(conv)
            conv = (
                LeakyReLU()(conv)
                if (conv_activation == "leaky_relu")
                else Activation(conv_activation)(conv)
            )
            if jj == num_conv_per_block - 1:  # if last conv in the block
                to_concat.append(conv)
                conv = MaxPooling2D(pool_size=(2, 2))(conv)
                if dor > 0:
                    conv = Dropout(dor)(conv)

    #    if(not(upsamp)): #8-km data
    #      nchan = input_tuples[3][-1]
    #      new_input = Input(shape=(None,None,nchan), name=f"input_{incter}"); incter+=1
    #      inputs.append(new_input)
    #      conv = concatenate([conv,new_input], axis=-1)

    # Middle
    for jj in range(num_conv_per_block):
        conv = Conv2D(
            nfmaps_by_block[-1],
            filter_width,
            padding=padding,
            dilation_rate=dilation,
            kernel_initializer=ki,
        )(conv)
        if batch_norm:
            conv = BatchNormalization()(conv)
        conv = (
            LeakyReLU()(conv)
            if (conv_activation == "leaky_relu")
            else Activation(conv_activation)(conv)
        )

    # decoding branch
    for ii in range(num_decoding_blocks):
        nfmaps = nfmaps_by_block[
            -ii - 1 - 1
        ]  # the extra -1 is because nfmaps_by_block[-1] is for the middle/bottleneck
        conv = Conv2DTranspose(
            nfmaps, filter_width, strides=2, padding=padding, kernel_initializer=ki
        )(conv)
        if not (noskips):
            conv = concatenate(
                [conv, to_concat[-ii - 1]], axis=-1
            )  # This adds the skip connections
        if dor > 0:
            conv = Dropout(dor)(conv)
        for jj in range(num_conv_per_block):
            conv = Conv2D(
                nfmaps,
                filter_width,
                padding=padding,
                dilation_rate=dilation,
                kernel_initializer=ki,
            )(conv)
            if batch_norm:
                conv = BatchNormalization()(conv)
            conv = (
                LeakyReLU()(conv)
                if (conv_activation == "leaky_relu")
                else Activation(conv_activation)(conv)
            )

    # set initial bias initialization
    output_bias = (
        tf.keras.initializers.Constant(config["bias_init"])
        if (config["bias_init"] is not None)
        else "zeros"
    )

    if loss_fcn == "QR":  # if quantile regression
        quantiles = [0.1, 0.25, 0.5, 0.75, 0.9]  # hard-coded
        quantile_outputs = []
        for _ in range(len(quantiles)):
            output = Conv2D(1, 1, activation="linear")(
                conv
            )  # Adjust channels based on your problem
            quantile_outputs.append(output)
        model = Model(inputs=inputs, outputs=quantile_outputs)
    else:  # normal classification
        conv = Conv2D(ntargets, (1, 1), bias_initializer=output_bias)(conv)
        if batch_norm:
            conv = BatchNormalization()(conv)
        conv = Activation("sigmoid")(conv)
        model = Model(inputs=inputs, outputs=conv)

    if loss_fcn == "QR":
        quantiles = [0.1, 0.25, 0.5, 0.75, 0.9]  # hard-coded
        loss_fcn = [
            lambda y_true, y_pred, q=q: losses.quantile_loss(q, y_true, y_pred)
            for q in quantiles
        ]

    optimizer = config["optimizer"]
    if optimizer.lower() == "adabelief":
        opt = tfa.optimizers.AdaBelief
    else:
        opt = getattr(tensorflow.keras.optimizers, config["optimizer"])

    METRICS = get_metrics()

    model.compile(
        optimizer=opt(learning_rate=learning_rate), loss=loss_fcn, metrics=METRICS
    )

    print(model.summary())

    return model


##############################################################################################################################
def convlstm2d(config):
    # build the CNN
    logging.info("Building the CNN...")
    input_tuples = config["input_tuples"]
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    batch_norm = config["batch_norm"]
    dor = config["dropout_rate"]
    dilation = config["dilation"]

    input_1 = Input((None, None, input_tuples[0][-1]))

    inputs = [input_1]
    concat_inputs = [input_1]

    if False and len(input_tuples) > 1:  # we have multiple resolutions
        upsamp_res = int(input_tuples[0][-2] / input_tuples[1][-2])
        input_2 = Input(
            shape=(None, None, input_tuples[1][-1])
        )  # Input(shape=input_tuples[1]) #Input(shape=(None,None,input_tuples[1][-1]))
        input_2_interp = UpSampling2D(
            size=(upsamp_res, upsamp_res), data_format="channels_last"
        )(input_2)
        inputs.append(input_2)
        concat_inputs.append(input_2_interp)
        if len(input_tuples) > 2:
            upsamp_res = int(input_tuples[0][-2] / input_tuples[2][-2])
            input_3 = Input(
                shape=(None, None, input_tuples[2][-1])
            )  # Input(shape=input_tuples[2]) #Input(shape=(None,None,input_tuples[2][-1]))
            input_3_interp = UpSampling2D(
                size=(upsamp_res, upsamp_res), data_format="channels_last"
            )(input_3)
            inputs.append(input_3)
            concat_inputs.append(input_3_interp)
            if (
                len(input_tuples) > 3
            ):  # ok, this assumes input1 is 0.5km, input2 is 1km, input3 is 2km, and input4 is SST, '4km'
                upsamp_res = int(input_tuples[0][-2] / input_tuples[3][-2])
                input_4 = Input(shape=input_tuples[3])
                input_4_interp = UpSampling2D(
                    size=(upsamp_res, upsamp_res), data_format="channels_last"
                )(input_4)
                inputs.append(input_4)
                concat_inputs.append(input_4_interp)

        final_input = concatenate(concat_inputs)
    else:
        final_input = input_1

    # input1 is 20 min ago, input2 is 10 min ago, input3 is 0 min ago....just 2km for now!
    input_2 = Input((None, None, input_tuples[1][-1]))
    input_3 = Input((None, None, input_tuples[2][-1]))
    inputs = [input_1, input_2, input_3]
    final_input = concatenate([input_1, input_2, input_3])

    def lag_branch(inp, num_conv_filters, filter_width, dilation_rate=dilation):
        for bb, factor in enumerate([1, 1.5, 2, 3]):  # num blocks = len of list here
            conv_in1 = inp if (bb == 0) else conv
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv_in1)
            conv = Activation(conv_activation)(conv)
            conv = Conv2D(
                num_conv_filters * factor,
                filter_width,
                padding="same",
                dilation_rate=dilation,
            )(conv)
            conv = Activation(conv_activation)(conv)
            conv = MaxPooling2D(pool_size=(2, 2))(conv)
        return conv

    # t - 2 branch
    t2 = lag_branch(input_1, num_conv_filters, filter_width, dilation_rate=dilation)

    # t - 1 branch
    t1 = lag_branch(input_2, num_conv_filters, filter_width, dilation_rate=dilation)

    # encoding branch3
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(input_3)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    conv1 = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv1)
    if batch_norm:
        conv1 = BatchNormalization()(conv1)
    conv1 = Activation(conv_activation)(conv1)
    pool1 = MaxPooling2D(pool_size=(2, 2))(conv1)
    if dor > 0:
        pool1 = Dropout(dor / 2.0)(pool1)

    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(pool1)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    conv2 = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv2)
    if batch_norm:
        conv2 = BatchNormalization()(conv2)
    conv2 = Activation(conv_activation)(conv2)
    pool2 = MaxPooling2D(pool_size=(2, 2))(conv2)
    if dor > 0:
        pool2 = Dropout(dor)(pool2)

    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(pool2)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    conv3 = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv3)
    if batch_norm:
        conv3 = BatchNormalization()(conv3)
    conv3 = Activation(conv_activation)(conv3)
    pool3 = MaxPooling2D(pool_size=(2, 2))(conv3)
    if dor > 0:
        pool3 = Dropout(dor)(pool3)

    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(pool3)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    conv4 = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv4)
    if batch_norm:
        conv4 = BatchNormalization()(conv4)
    conv4 = Activation(conv_activation)(conv4)
    pool4 = MaxPooling2D(pool_size=(2, 2))(conv4)
    if dor > 0:
        pool4 = Dropout(dor)(pool4)

    # make the time dim and concatenate along that dim
    t1 = tf.expand_dims(t1, 1)
    t2 = tf.expand_dims(t2, 1)
    pool4 = tf.expand_dims(pool4, 1)

    concat = concatenate([t1, t2, pool4], axis=1)

    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(concat)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(conv)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=True,
    )(conv)
    conv = ConvLSTM2D(
        filters=num_conv_filters * 3,
        kernel_size=(filter_width, filter_width),
        padding="same",
        return_sequences=False,
    )(conv)

    up = Conv2DTranspose(num_conv_filters * 3, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv4], axis=-1)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 3, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 2, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv3], axis=-1)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 2, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(
        num_conv_filters * 1.5, filter_width, strides=2, padding="same"
    )(
        conv
    )  # no activation here?
    merge = concatenate([up, conv2], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1.5, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    up = Conv2DTranspose(num_conv_filters * 1, filter_width, strides=2, padding="same")(
        conv
    )  # no activation here?
    merge = concatenate([up, conv1], axis=-1)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(merge)
    conv = Activation(conv_activation)(conv)
    conv = Conv2D(
        num_conv_filters * 1, filter_width, padding="same", dilation_rate=dilation
    )(conv)
    conv = Activation(conv_activation)(conv)

    # set initial bias initialization
    output_bias = (
        tf.keras.initializers.Constant(config["bias_init"])
        if (config["bias_init"] is not None)
        else "zeros"
    )

    conv = Conv2D(1, (1, 1), bias_initializer=output_bias)(conv)
    conv = Activation("sigmoid")(conv)
    if "CH02" in config["inputs"][0]:  # reduce "500m" to "2km"
        conv = MaxPooling2D(pool_size=(4, 4))(conv)
    elif "CH05" in config["inputs"][0]:  # reduce "1km" to "2km"
        conv = MaxPooling2D(pool_size=(2, 2))(conv)

    # remove superfluous dims
    conv = tf.keras.backend.squeeze(conv, -1)

    model = Model(inputs=inputs, outputs=conv)

    loss_fcn = config["loss_fcn"]
    opt = getattr(tensorflow.keras.optimizers, config["optimizer"])

    model.compile(optimizer=opt(lr=learning_rate), loss=loss_fcn)

    print(model.summary())

    return model


##############################################################################################################################
def create_cnn_2km_VGG16(input_tuples, config, **kwargs):
    # build the CNN
    logging.info("Building the CNN...")
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    l2_param = config["l2_param"]
    dor = config["dropout_rate"]

    from tensorflow.keras.applications import VGG16

    conv_base = VGG16(
        weights="imagenet", include_top=False, input_shape=input_tuples[0]
    )

    # Flatten the last convolutional layer into a long feature vector
    conv_net = Flatten()(conv_base.output)

    # Dense layer 1
    conv_net = Dense(128)(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    # Dropout
    conv_net = Dropout(dor)(conv_net)

    # Dense layer 2
    conv_net = Dense(16)(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    # Dropout
    conv_net = Dropout(dor)(conv_net)

    # Dense output layer, equivalent to a logistic regression on the last layer
    conv_net = Dense(1)(conv_net)
    conv_net = Activation("sigmoid")(conv_net)
    conv_model = Model(inputs=[conv_in], outputs=conv_net)
    # Use the Adam optimizer with default parameters
    opt = Adam(lr=learning_rate)
    conv_model.compile(opt, "binary_crossentropy", metrics=LIST_OF_METRIC_FUNCTIONS)
    print(conv_model.summary())

    return conv_model


###################################################################################################
def create_cnn_3D(input_tuples, config, **kwargs):
    # build the CNN
    logging.info("Building the CNN...")
    num_conv_filters = config["num_conv_filters"]
    filter_width = config["filter_width"]
    conv_activation = config["conv_activation"]
    learning_rate = config["learning_rate"]
    l2_param = config["l2_param"]
    dor = config["dropout_rate"]
    scalar_vars = config["scalar_vars"]
    batch_norm = config["batch_norm"]
    nframes = config["nframes"]
    n2d_inputs = len(input_tuples) - 1 if (len(scalar_vars) > 1) else len(input_tuples)

    input_1 = Input(shape=input_tuples[0])
    inputs = [input_1]

    if len(input_tuples) > 1:  # we have multiple resolutions
        input_2 = Input(shape=input_tuples[1])
        input_2_interp = UpSampling3D(size=(1, 4, 4), data_format="channels_last")(
            input_2
        )
        final_input = concatenate([input_1, input_2_interp])
        inputs.append(input_2)
    else:
        final_input = input_1

    # First 3D convolution Layer
    conv_net = Conv3D(
        num_conv_filters, (filter_width, filter_width, filter_width), padding="same"
    )(final_input)
    conv_net = Activation(conv_activation)(conv_net)
    if batch_norm:
        conv_net = BatchNormalization()(conv_net)
    conv_net = MaxPooling3D(pool_size=(1, 2, 2))(conv_net)

    conv_net = Conv3D(
        num_conv_filters * 2, (filter_width, filter_width, filter_width), padding="same"
    )(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    if batch_norm:
        conv_net = BatchNormalization()(conv_net)
    conv_net = MaxPooling3D(pool_size=(1, 2, 2))(conv_net)

    conv_net = Conv3D(
        num_conv_filters * 4, (filter_width, filter_width, filter_width), padding="same"
    )(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    if batch_norm:
        conv_net = BatchNormalization()(conv_net)
    conv_net = MaxPooling3D(pool_size=(1, 2, 2))(conv_net)

    # Reshape to 2D
    # conv_net = tensorflow.keras.layers.Reshape((32,32,16))(conv_net)

    conv_net = Conv3D(
        num_conv_filters * 8, (filter_width, filter_width, filter_width), padding="same"
    )(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    if batch_norm:
        conv_net = BatchNormalization()(conv_net)
    conv_net = MaxPooling3D(pool_size=(2, 2, 2))(conv_net)

    conv_net = Conv3D(
        num_conv_filters * 16,
        (filter_width, filter_width, filter_width),
        padding="same",
    )(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    if batch_norm:
        conv_net = BatchNormalization()(conv_net)
    conv_net = MaxPooling3D(pool_size=(2, 2, 2))(conv_net)

    conv_net = Flatten()(conv_net)

    if scalar_vars:
        # add scalars here (e.g., lat, lon, SST, mlcape)
        scalar_input = Input(shape=input_tuples[-1])
        conv_net = concatenate([conv_net, scalar_input])
        inputs.append(scalar_input)

    # Dense layer 1
    conv_net = Dense(128)(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    conv_net = Dropout(dor)(conv_net)

    # Dense layer 2
    conv_net = Dense(16)(conv_net)
    conv_net = Activation(conv_activation)(conv_net)
    conv_net = Dropout(dor)(conv_net)

    # Dense output layer, equivalent to a logistic regression on the last layer
    conv_net = Dense(1)(conv_net)
    conv_net = Activation("sigmoid")(conv_net)

    conv_model = Model(inputs=inputs, outputs=conv_net)

    # opt = RAdam(learning_rate=learning_rate)
    opt = Adam(lr=learning_rate)
    conv_model.compile(opt, "binary_crossentropy", metrics=LIST_OF_METRIC_FUNCTIONS)
    print(conv_model.summary())

    return conv_model


###################################################################################################
